# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Functions for applying actuators to a physics pipeline."""

from typing import Tuple

from brax.v2 import kinematics
from brax.v2 import scan
from brax.v2.base import math
from brax.v2.base import Motion, System, Transform
import jax
from jax import numpy as jp


def to_tau(sys: System, act: jp.ndarray, q: jp.ndarray) -> jp.ndarray:
  """Convert actuator to a joint force tau.

  Args:
    sys: system defining the kinematic tree and other properties
    act: (act_size,) actuator force input vector
    q: joint position vector

  Returns:
    tau: (qd_size,) vector of joint forces
  """
  if sys.act_size() == 0:
    return jp.zeros(sys.qd_size())

  def act_fn(act_type, act, actuator, q, qd_idx):
    if act_type not in ('p', 'm'):
      raise RuntimeError(f'unrecognized act type: {act_type}')

    force = jp.clip(act, actuator.ctrl_range[:, 0], actuator.ctrl_range[:, 1])
    if act_type == 'p':
      force -= q  # positional actuators have a bias
    tau = actuator.gear * force

    return tau, qd_idx

  qd_idx = jp.arange(sys.qd_size())
  tau, qd_idx = scan.actuator_types(
      sys, act_fn, 'aaqd', 'a', act, sys.actuator, q, qd_idx
  )
  tau = jp.zeros(sys.qd_size()).at[qd_idx].add(tau)

  return tau


def to_tau_world(
    sys: System, q: jp.ndarray, tau: jp.ndarray
) -> Tuple[Motion, jp.ndarray, jp.ndarray]:
  """Converts joint force tau to world frame.

  Args:
    sys: system defining the kinematic tree and other properties
    q: joint position vector
    tau: joint force

  Returns:
    f_a: joint force from actuator in world frame
    pos: position in world space the actuator is being applied
    f_idxs: indices where actuators are acting
  """

  # convert joint position/velocity to transform/motion in joint frame
  def jcalc(typ, q, tau, motion):
    if typ == 'f':
      q, tau = q.reshape((-1, 7)), jp.zeros((q.shape[0], 6))
      j = Transform(pos=q[:, 0:3], rot=q[:, 3:7])
      # no actuators for free joints
      tau = Motion(ang=jp.zeros_like(j.pos), vel=jp.zeros_like(j.pos))
    else:
      # create joint transforms and motions:
      # - rotation/velocity about axis for revolute (motion.ang)
      # - translation/velocity along axis for prismatic (motion.vel)
      rot_fn = lambda ang, q: math.normalize(math.quat_rot_axis(ang, q))[0]
      j = Transform.create(
          rot=jax.vmap(rot_fn)(motion.ang, q),
          pos=jax.vmap(jp.multiply)(motion.vel, q),
      )
      tau = jax.vmap(lambda a, b: a * b)(motion, tau)

      # then group them by link, so each link has num_dofs joints
      num_links, num_dofs = tau.ang.shape[0] // int(typ), int(typ)
      s = (num_links, num_dofs, -1)
      j_stack, tau_stack = j.reshape(s), tau.reshape(s)

      # accumulate j and tau one dof at a time
      j, tau = j_stack.take(0, axis=1), tau_stack.take(0, axis=1)
      for i in range(1, num_dofs):
        j_i, tau_i = j_stack.take(i, axis=1), tau_stack.take(i, axis=1)
        j = j.vmap().do(j_i)
        tau = tau + Motion(
            ang=jax.vmap(math.rotate)(tau_i.ang, j_i.rot),
            vel=jax.vmap(math.rotate)(tau_i.vel, j_i.rot),
        )

    return tau

  p_idx = jp.array(sys.link_parents)
  c_idx = jp.array(range(sys.num_links()))

  x, _ = kinematics.forward(sys, q, jp.zeros(sys.qd_size()))
  x_pad = jax.tree_map(lambda x, y: jp.vstack((x, y)), x, Transform.zero((1,)))
  x_p = x_pad.take(p_idx)
  x_c = x.vmap().do(sys.link.joint)
  x_joint = x_p.vmap().do(sys.link.transform).vmap().do(sys.link.joint)

  f_a = scan.link_types(sys, jcalc, 'qdd', 'l', q, tau, sys.dof.motion)

  f_a = jax.tree_map(lambda x, y: jp.vstack([x, y]), f_a, -f_a)
  pos = jp.vstack((x_c.pos, x_joint.pos))
  link_idx = jp.hstack((c_idx, p_idx))
  f_a *= link_idx.reshape((-1, 1)) != -1

  return f_a, pos, link_idx



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""A brax environment for training and inference."""

import abc
from typing import Any, Dict

from brax.v2 import base
from brax.v2.generalized import pipeline as g_pipeline
from brax.v2.spring import pipeline as s_pipeline
from flax import struct
import jax
from jax import numpy as jp


@struct.dataclass
class State:
  """Environment state for training and inference."""

  pipeline_state: base.State
  obs: jp.ndarray
  reward: jp.ndarray
  done: jp.ndarray
  metrics: Dict[str, jp.ndarray] = struct.field(default_factory=dict)
  info: Dict[str, Any] = struct.field(default_factory=dict)


class Env(abc.ABC):
  """Interface for driving training and inference."""

  @abc.abstractmethod
  def reset(self, rng: jp.ndarray) -> State:
    """Resets the environment to an initial state."""

  @abc.abstractmethod
  def step(self, state: State, action: jp.ndarray) -> State:
    """Run one timestep of the environment's dynamics."""

  @property
  @abc.abstractmethod
  def observation_size(self) -> int:
    """The size of the observation vector returned in step and reset."""

  @property
  @abc.abstractmethod
  def action_size(self) -> int:
    """The size of the action vector expected by step."""

  @property
  @abc.abstractmethod
  def backend(self) -> str:
    """The physics backend that this env was instatiated with."""

  @property
  def unwrapped(self) -> 'Env':
    return self


class PipelineEnv(Env):
  """API for driving a brax system for training and inference."""

  __pytree_ignore__ = (
      '_backend',
      '_pipeline',
  )

  def __init__(
      self, sys: base.System, backend: str = 'generalized', n_frames: int = 1
  ):
    """Initializes PipelineEnv.

    Args:
      sys: system defining the kinematic tree and other properties
      backend: string specifying the physics pipeline
      n_frames: the number of times to step the physics pipeline for each
        environment step
    """
    self.sys = sys

    pipeline = {'generalized': g_pipeline, 'spring': s_pipeline}
    if backend not in pipeline:
      raise ValueError(f'backend should be in {pipeline.keys()}.')

    self._backend = backend
    self._pipeline = pipeline[backend]
    self._n_frames = n_frames

  def pipeline_step(
      self, pipeline_state: Any, action: jp.ndarray
  ) -> base.State:
    """Takes a physics step using the physics pipeline."""
    def f(state, _):
      return self._pipeline.step(self.sys, state, action), None

    return jax.lax.scan(f, pipeline_state, (), self._n_frames)[0]

  @property
  def dt(self) -> jp.ndarray:
    """The timestep used for each env step."""
    return self.sys.dt * self._n_frames

  @property
  def observation_size(self) -> int:
    rng = jax.random.PRNGKey(0)
    reset_state = self.unwrapped.reset(rng)
    return reset_state.obs.shape[-1]

  @property
  def action_size(self) -> int:
    return self.sys.act_size()

  @property
  def backend(self) -> str:
    return self._backend


class Wrapper(Env):
  """Wraps an environment to allow modular transformations."""

  def __init__(self, env: Env):
    self.env = env

  def reset(self, rng: jp.ndarray) -> State:
    return self.env.reset(rng)

  def step(self, state: State, action: jp.ndarray) -> State:
    return self.env.step(state, action)

  @property
  def observation_size(self) -> int:
    return self.env.observation_size

  @property
  def action_size(self) -> int:
    return self.env.action_size

  @property
  def unwrapped(self) -> Env:
    return self.env.unwrapped

  @property
  def backend(self) -> str:
    return self.unwrapped.backend

  def __getattr__(self, name):
    if name == '__setstate__':
      raise AttributeError(name)
    return getattr(self.env, name)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Observation descriptions."""
from typing import Union, Tuple, Any

OBS_INDICES = dict(
    vel=dict(
        ant=(13, 14),
        humanoid=(22, 23),
        halfcheetah=(11,),
        hopper=(8,),
        walker2d=(11,),
        uni_ant=(('body_vel:torso_ant1', 0), ('body_vel:torso_ant1', 1)),
        bi_ant=(('body_vel:torso_ant1', 0), ('body_vel:torso_ant2', 0)),
    ),)


def register_indices(env_name: str, indices_type: str,
                     indices: Tuple[Union[int, Any]]):
  """Register indices."""
  global OBS_INDICES
  if indices_type not in OBS_INDICES:
    OBS_INDICES[indices_type] = {}
  OBS_INDICES[indices_type][env_name] = indices


def get_indices(env_name: str, indices_type: str):
  """Get indices."""
  return OBS_INDICES.get(indices_type, {}).get(env_name, None)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evolution strategy networks."""

from typing import Sequence, Tuple

from brax.training import distribution
from brax.training import networks
from brax.training import types
from brax.training.types import PRNGKey
import flax
from flax import linen


@flax.struct.dataclass
class ESNetworks:
  policy_network: networks.FeedForwardNetwork
  parametric_action_distribution: distribution.ParametricDistribution


def make_inference_fn(es_networks: ESNetworks):
  """Creates params and inference function for the ES agent."""

  def make_policy(params: types.PolicyParams,
                  deterministic: bool = False) -> types.Policy:

    def policy(observations: types.Observation,
               key_sample: PRNGKey) -> Tuple[types.Action, types.Extra]:
      logits = es_networks.policy_network.apply(*params, observations)
      if deterministic:
        return es_networks.parametric_action_distribution.mode(logits), {}
      return es_networks.parametric_action_distribution.sample(
          logits, key_sample), {}

    return policy

  return make_policy


def make_es_networks(
    observation_size: int,
    action_size: int,
    preprocess_observations_fn: types.PreprocessObservationFn = types
    .identity_observation_preprocessor,
    hidden_layer_sizes: Sequence[int] = (32,) * 4,
    activation: networks.ActivationFn = linen.relu) -> ESNetworks:
  """Make ES networks."""
  parametric_action_distribution = distribution.NormalTanhDistribution(
      event_size=action_size)
  policy_network = networks.make_policy_network(
      parametric_action_distribution.param_size,
      observation_size,
      preprocess_observations_fn=preprocess_observations_fn,
      hidden_layer_sizes=hidden_layer_sizes,
      activation=activation)
  return ESNetworks(
      policy_network=policy_network,
      parametric_action_distribution=parametric_action_distribution)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brax.envs."""

import logging
import time

from absl.testing import absltest
from absl.testing import parameterized
from brax import envs
from brax import jumpy as jp
import jax

_EARLY_TERMINATION = ('ant', 'humanoid')
_EXPECTED_SPS = {'ant': 1000, 'fetch': 1000}


class EnvTest(parameterized.TestCase):

  @parameterized.parameters(_EXPECTED_SPS.items())
  def testSpeed(self, env_name, expected_sps):
    batch_size = 128
    episode_length = 1000
    auto_reset = (env_name not in _EARLY_TERMINATION)

    env = envs.create(
        env_name,
        episode_length=episode_length,
        auto_reset=auto_reset,
        batch_size=batch_size)
    zero_action = jp.zeros((batch_size, env.action_size))

    @jax.jit
    def run_env(state):

      def step(carry, _):
        state, = carry
        state = env.step(state, zero_action)
        return (state,), ()

      (state,), _ = jax.lax.scan(step, (state,), (), length=episode_length)
      return state

    # warmup
    rng = jp.random_prngkey(0)
    state = jax.jit(env.reset)(rng)
    state = run_env(state)
    state.done.block_until_ready()

    sps = []
    for seed in range(5):
      rng = jp.random_prngkey(seed)
      state = jax.jit(env.reset)(rng)
      jax.device_put(state)
      t = time.time()
      state = run_env(state)
      state.done.block_until_ready()
      sps.append((batch_size * episode_length) / (time.time() - t))
      self.assertTrue(jp.all(state.done))

    mean_sps = jp.mean(jp.array(sps))
    logging.info('%s SPS %s %s', env_name, mean_sps, sps)
    self.assertGreater(mean_sps, expected_sps * 0.99)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluators for VGRL."""


import copy
import functools
import itertools
from typing import Dict, Tuple, Any
from brax.experimental.braxlines.common import evaluators
from brax.experimental.braxlines.vgcrl.utils import Discriminator
from brax.experimental.composer import observers
from brax.io import file
from brax.io import html
import jax
from jax import numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import tensorflow_probability as tfp

tfp = tfp.substrates.jax
tfd = tfp.distributions


@jax.jit
def jit_compute_distance(env_vars: jnp.ndarray, goals: jnp.ndarray):
  dist_vector = jnp.mean(jnp.abs(env_vars - goals[None, None]), axis=(0, 1, 2))
  dist = jnp.mean(
      jnp.sqrt(jnp.mean((env_vars - goals[None, None])**2, axis=-1)))
  return dist_vector, dist


@functools.partial(jax.jit, static_argnums=(2,))
def jit_compute_mi_1d(env_vars: jnp.ndarray, obs_scale: jnp.ndarray,
                      num_1d_bins: int):
  """Use binning to compute 1d mutual information."""
  hist = jnp.histogram(
      env_vars.flatten(), bins=num_1d_bins, range=(-obs_scale, obs_scale))
  discretization_all = hist[0] / env_vars.flatten().shape[0]
  entropy_all = -jnp.sum(
      discretization_all * jnp.log(discretization_all + 1e-12))
  discretization_z = [
      jnp.histogram(env_vars[..., j].flatten(), bins=hist[1])[0] /
      env_vars[..., j].flatten().shape[0] for j in range(env_vars.shape[-1])
  ]
  entropy_z = -jnp.mean(
      jnp.array(
          [jnp.sum(p_z * jnp.log(p_z + 1e-12)) for p_z in discretization_z]))
  mi = entropy_all - entropy_z
  return mi, entropy_all, entropy_z


def estimate_latent_goal_reaching_metric(disc: Discriminator,
                                         params: Dict[str, Dict[str,
                                                                jnp.ndarray]],
                                         goals: jnp.ndarray,
                                         time_subsampling: int = 1,
                                         time_last_n: int = 500,
                                         verbose: bool = False,
                                         **kwargs):
  """Compute latent-goal reaching metric."""
  assert not disc.normalize_obs
  assert goals.shape[
      -1] == disc.indexed_obs_size, f'{goals.shape}[-1] != {disc.indexed_obs_size}'
  _, I = goals.shape

  # TODO: add stochastic q(z) sampling
  goal_obs = disc.unindex_obs(goals)
  dist_q = disc.dist_q_z_o(goal_obs, params['extra'])
  if isinstance(dist_q, tfd.MultivariateNormalDiag):
    zs = dist_q.loc
  elif isinstance(dist_q, tfd.OneHotCategorical):
    zs = jax.nn.one_hot(jnp.argmax(dist_q.logits, axis=-1), disc.z_size)
  else:
    raise NotImplementedError(dist_q)

  _, _, _, _, _, env_vars = rollout_skills(
      params=params, disc=disc, verbose=verbose, num_z=None, zs=zs, **kwargs)
  env_vars = env_vars[-time_last_n:][::time_subsampling]  # [T, M, D, I]
  dist_vector, dist = jit_compute_distance(env_vars, goals)
  metrics = {
      f'metrics/lgr_1d/{disc.obs_labels[i]}': dist_vector[i] for i in range(I)
  }
  metrics.update({
      'metrics/lgr': dist,
  })
  return metrics


def estimate_empowerment_metric(disc: Discriminator,
                                time_subsampling: int = 1,
                                time_last_n: int = 500,
                                num_1d_bins: int = 1000,
                                num_2d_bins: int = 1000,
                                verbose: bool = True,
                                include_1d: bool = True,
                                include_2d: bool = False,
                                custom_obs_indices: Tuple[Any] = (),
                                custom_obs_scale: float = None,
                                **kwargs):
  """Compute non-parametric estimate of mutual information."""
  env, _, zs, _, env_obses, _ = rollout_skills(
      disc=disc, verbose=verbose, **kwargs)

  obs_indices = copy.deepcopy(disc.obs_indices)
  obs_scale = copy.deepcopy(disc.obs_scale)
  obs_labels = copy.deepcopy(disc.obs_labels)
  if custom_obs_indices:
    custom_obs_indices, custom_obs_labels = observers.index_preprocess(
        custom_obs_indices, env)
    custom_indexed_obs_size = len(custom_obs_indices)
    custom_obs_scale = jnp.array(custom_obs_scale or
                                 1.) * jnp.ones(custom_indexed_obs_size)
    obs_labels += custom_obs_labels
    obs_indices += custom_obs_indices
    obs_scale = jnp.concatenate([obs_scale, custom_obs_scale])
  env_vars = env_obses[..., obs_indices]

  D = zs.shape[0]
  I = env_vars.shape[-1]

  env_vars = env_vars[-time_last_n:][::time_subsampling]  # [T, M, D, I]

  metrics = {}

  # compute 1D estimates
  if include_1d:
    for i in range(I):
      mi, entropy_all, entropy_z = jit_compute_mi_1d(env_vars[..., i],
                                                     obs_scale[i], num_1d_bins)
      metrics.update({
          f'metrics/mi_1d/{obs_labels[i]}': mi,
          f'metrics/entropy_all_1d/{obs_labels[i]}': entropy_all,
          f'metrics/entropy_z_1d/{obs_labels[i]}': entropy_z,
      })

  # compute 2D estimates
  # jax histogram2d has an error. Switch to np for now.
  if include_2d:
    env_vars = np.array(env_vars)
    obs_scale = np.array(obs_scale)
    for i1, i2 in itertools.combinations(range(I), 2):
      if i1 == i2:
        continue
      hist = np.histogram2d(
          env_vars[..., i1].flatten(),
          env_vars[..., i2].flatten(),
          bins=num_2d_bins,
          range=((-obs_scale[i1], obs_scale[i1]), (-obs_scale[i2],
                                                   obs_scale[i2])))
      etization_all = hist[0] / env_vars[..., i1].flatten().shape[0]
      entropy_all = -np.sum(etization_all * np.log(etization_all + 1e-12))
      etization_z = [
          np.histogram2d(
              env_vars[..., j, i1].flatten(),
              env_vars[..., j, i2].flatten(),
              bins=hist[1:])[0] / env_vars[..., j, i1].flatten().shape[0]
          for j in range(D)
      ]
      entropy_z = -np.mean(
          np.array([np.sum(p_z * np.log(p_z + 1e-12)) for p_z in etization_z]))
      mi = entropy_all - entropy_z
      metrics.update({
          f'metrics/mi_2d/{obs_labels[i1]}_{obs_labels[i2]}':
              mi,
          f'metrics/entropy_all_2d/{obs_labels[i1]}_{obs_labels[i2]}':
              entropy_all,
          f'metrics/entropy_z_2d/{obs_labels[i1]}_{obs_labels[i2]}':
              entropy_z,
      })

  return metrics


def rollout_skills(
    params: Dict[str, Dict[str, jnp.ndarray]],
    env_fn,
    disc: Discriminator,
    inference_fn=None,
    verbose: bool = True,
    num_z: int = 10,
    num_samples_per_z: int = 10,
    seed: int = 0,
    zs: jnp.ndarray = None,
):
  """Rollout skills in parallel."""
  env = env_fn()
  O = disc.env_obs_size
  I = disc.indexed_obs_size
  Z = disc.z_size
  M = num_samples_per_z

  rng = jax.random.PRNGKey(seed)
  if zs is None:
    # Sample {D} z's [D, Z]
    dist_p = disc.dist_p_z()
    if isinstance(dist_p, tfd.Deterministic):
      zs = jnp.array([dist_p.loc])  # D=1
    elif isinstance(dist_p, tfd.OneHotCategorical):
      zs = jax.nn.one_hot(jnp.arange(0, disc.z_size), disc.z_size)  # D=Z
    else:
      rng, z_key = jax.random.split(rng)
      zs = disc.sample_p_z(num_z, z_key)  # D=num_z
  D = zs.shape[0]

  # Repeat each z by {M} times
  batch_z = jnp.repeat(zs, M, axis=0)  # [D*M, Z]

  batch_env, states = evaluators.rollout_env(
      env_fn=env_fn,
      params=params,
      batch_size=D * M,
      inference_fn=inference_fn,
      reset_args=(batch_z,),
      step_args=() if params is None else
      (params['normalizer'], params['extra']),
      step_fn_name='step2' if params is None else 'step',
  )

  # Get env dimensions of interest
  obses = jnp.stack([state.obs for state in states], axis=0)  # [T, D*M, O+D]
  if verbose:
    print((f'T={len(states)}, O={O}, Z={Z}, D={D}, M={M}'))
  env_obses, _ = batch_env.disc.split_obs(obses)  # [T, D*M, O]
  env_vars = batch_env.disc.index_obs(env_obses)  # [T, D*M, I]
  env_obses = env_obses.reshape(-1, D, M, O).swapaxes(1, 2)  # [T, M, D, O]
  env_vars = env_vars.reshape(-1, D, M, I).swapaxes(1, 2)  # [T, M, D, I]
  return env, batch_env, zs, states, env_obses, env_vars


def visualize_skills(output_path: str = None,
                     time_subsampling: int = 10,
                     time_last_n: int = 500,
                     output_name: str = 'skills',
                     save_video: bool = False,
                     disc: Discriminator = None,
                     num_samples_per_z: int = 5,
                     **kwargs):
  """Visualizing skills in a 2D plot."""
  env, _, zs, states, env_obses, env_vars = rollout_skills(
      disc=disc, num_samples_per_z=num_samples_per_z, **kwargs)
  D = zs.shape[0]
  I = env_vars.shape[-1]

  env_obses = env_obses[-time_last_n:][::time_subsampling]
  env_vars = env_vars[-time_last_n:][::time_subsampling]

  labels = disc.obs_labels
  assert env_vars.shape[-1] == len(
      labels), f'{env_vars.shape}[-1] != {len(labels)}'
  if env_vars.shape[-1] % 2:
    env_vars = jnp.concatenate(
        [env_vars, jnp.zeros(env_vars.shape[:-1] + (1,))], axis=-1)
    labels = tuple(labels) + (None,)
  env_vars = env_vars.reshape(-1, D, I)  # [T*M, D, I]

  # Plot
  def spec(N):
    t = np.linspace(-510, 510, N)
    return np.clip(np.stack([-t, 510 - np.abs(t), t], axis=1), 0, 255) / 255.

  colours = spec(D)  # [D, 3]
  nfigs = int(len(labels) / 2)
  _, axs = plt.subplots(ncols=nfigs, figsize=(7 * nfigs, 6))
  if nfigs == 1:
    axs = [axs]
  for i in range(nfigs):
    for j, (z, c) in enumerate(zip(zs, colours)):
      axs[i].scatter(
          x=env_vars[:, j, 2 * i],
          y=env_vars[:, j, 2 * i + 1],
          c=c,
          alpha=0.3,
          label=f'z={z}')
      axs[i].legend()
      axs[i].set(xlabel=f'{labels[2*i]}', ylabel=f'{labels[2*i+1]}')
  if output_path:
    file.MakeDirs(output_path)
    with file.File(f'{output_path}/{output_name}.png', 'wb') as f:
      plt.savefig(f)
    with file.File(f'{output_path}/{output_name}.npy', 'wb') as f:
      np.save(f, zs)
    if save_video:

      def index_fn(x, i):
        return x[i]

      for i in range(D):
        html.save_html(
            f'{output_path}/{output_name}_skill{i:02}.html',
            env.sys, [
                jax.tree_util.tree_map(
                    functools.partial(index_fn, i=i * num_samples_per_z),
                    state.qp) for state in states
            ],
            make_dir=True)
  return states



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Saves a system config and trajectory as json."""

import json
from typing import List, Optional, Text

import brax
from brax.io.file import File
import jax.numpy as jnp
import numpy as onp

from google.protobuf import json_format


class JaxEncoder(json.JSONEncoder):

  def default(self, obj):
    if isinstance(obj, jnp.ndarray):
      return obj.tolist()
    if isinstance(obj, onp.ndarray):
      return obj.tolist()
    return json.JSONEncoder.default(self, obj)


def dumps(sys: brax.System,
          qps: List[brax.QP],
          info: Optional[List[brax.Info]] = None) -> Text:
  """Creates a json string of the system config."""
  d = {
      'config': json_format.MessageToDict(sys.config, True),
      'pos': [qp.pos for qp in qps],
      'rot': [qp.rot for qp in qps],
      'debug': info is not None,
  }
  if info:
    # Add debug info for the contact points.
    max_len = max([sum(onp.array(i.contact_penetration) > 0) for i in info])

    def _pad_arr(arr):
      arr = onp.array(arr)
      padding = -onp.ones(max_len - arr.shape[0])
      if len(arr.shape) > 1:
        padding = -onp.ones((max_len - arr.shape[0], arr.shape[1]))
      return onp.concatenate([arr, padding])

    # Pad the contact points.
    d['contact_pos'] = [
        _pad_arr(i.contact_pos[i.contact_penetration > 0]) for i in info
    ]
    d['contact_normal'] = [
        _pad_arr(i.contact_normal[i.contact_penetration > 0]) for i in info
    ]
    d['contact_penetration'] = [
        _pad_arr(i.contact_penetration[i.contact_penetration > 0]) for i in info
    ]
  return json.dumps(d, cls=JaxEncoder)


def save(path: str,
         sys: brax.System,
         qps: List[brax.QP],
         info: Optional[List[brax.Info]] = None):
  """Saves a system config and trajectory as json."""
  with File(path, 'w') as fout:
    system = dumps(sys, qps, info)
    fout.write(system)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PPO tests."""
import pickle

from absl.testing import absltest
from absl.testing import parameterized
from brax import envs
from brax.training.acme import running_statistics
from brax.training.agents.ppo import networks as ppo_networks
from brax.training.agents.ppo import train as ppo
import jax


class PPOTest(parameterized.TestCase):
  """Tests for PPO module."""


  def testTrain(self):
    """Test PPO with a simple env."""
    _, _, metrics = ppo.train(
        envs.get_environment('fast'),
        num_timesteps=2**15,
        episode_length=128,
        num_envs=64,
        learning_rate=3e-4,
        entropy_cost=1e-2,
        discounting=0.95,
        unroll_length=5,
        batch_size=64,
        num_minibatches=8,
        num_updates_per_batch=4,
        normalize_observations=True,
        seed=2,
        reward_scaling=10,
        normalize_advantage=False)
    self.assertGreater(metrics['eval/episode_reward'], 135)

  @parameterized.parameters(True, False)
  def testNetworkEncoding(self, normalize_observations):
    env = envs.get_environment('fast')
    original_inference, params, _ = ppo.train(
        env,
        num_timesteps=128,
        episode_length=128,
        num_envs=128,
        normalize_observations=normalize_observations)
    normalize_fn = lambda x, y: x
    if normalize_observations:
      normalize_fn = running_statistics.normalize
    ppo_network = ppo_networks.make_ppo_networks(env.observation_size,
                                                 env.action_size, normalize_fn)
    inference = ppo_networks.make_inference_fn(ppo_network)
    byte_encoding = pickle.dumps(params)
    decoded_params = pickle.loads(byte_encoding)

    # Compute one action.
    state = env.reset(jax.random.PRNGKey(0))
    original_action = original_inference(decoded_params)(
        state.obs, jax.random.PRNGKey(0))[0]
    action = inference(decoded_params)(state.obs, jax.random.PRNGKey(0))[0]
    self.assertSequenceEqual(original_action, action)
    env.step(state, action)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example: an existing Env + a new reward."""
AUTHORS = ('Hiroki Furuta', 'Shixiang Shane Gu')
CONTACTS = ('furuta@weblab.t.u-tokyo.ac.jp', 'shanegu@google.com')
AFFILIATIONS = ('u-tokyo.ac.jp', 'google.com')
DESCRIPTIONS = (
    'HalfCheetah running and jumping',
    'a cheetah hopping',
)

ENVS = dict(
    cheetah=dict(
        module='cheetah:JumpCheetah',
        tracks=('rl',),
    ),)

COMPONENTS = None



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Distribution utilility functions."""

from jax import numpy as jnp
import tensorflow_probability as tfp

tfp = tfp.substrates.jax
tfd = tfp.distributions


def clipped_onehot_categorical(logits: jnp.ndarray, clip_range: float = 0):
  if clip_range:
    assert clip_range > 0.0, clip_range
    logits -= jnp.max(logits, axis=-1, keepdims=True)
    logits = jnp.clip(logits, a_min=-clip_range)
  return tfd.OneHotCategorical(logits=logits)


def clipped_bernoulli(logits: jnp.ndarray, clip_range: float = 0):
  if clip_range:
    assert clip_range > 0.0, clip_range
    logits = jnp.clip(logits, a_min=-clip_range, a_max=clip_range)
  return tfd.Bernoulli(logits=logits)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Useful functions for creating and processing meshes."""

import itertools

from brax.v2.base import Box, Mesh
from jax import numpy as jp


_BOX_CORNERS = list(itertools.product((-1, 1), (-1, 1), (-1, 1)))

# pyformat: disable
# The faces of a triangulated box, i.e. the indices in _BOX_CORNERS of the
# vertices of the 12 triangles (two triangles for each side of the box).
_TRIANGULATED_BOX_FACES = [
    0, 1, 4, 4, 1, 5,  # front
    0, 4, 2, 2, 4, 6,  # bottom
    6, 4, 5, 6, 5, 7,  # right
    2, 6, 3, 3, 6, 7,  # back
    1, 3, 5, 5, 3, 7,  # top
    0, 2, 1, 1, 2, 3,  # left
]
# pyformat: enable


def box(b: Box) -> Mesh:
  """Creates a mesh from a box geometry."""
  assert len(b.halfsize.shape) == 2 and b.halfsize.shape[-1] == 3, (
      'Box halfsize should have a batch dimension and have length 3, '
      f'got {b.halfsize.shape}'
  )
  n_boxes = b.halfsize.shape[0]
  box_corners = jp.array([_BOX_CORNERS] * n_boxes)
  vert = box_corners * b.halfsize.reshape(n_boxes, -1, 3)
  face = jp.array([_TRIANGULATED_BOX_FACES] * n_boxes).reshape(n_boxes, -1, 3)
  return Mesh(
      face=face,
      vert=vert,
      link_idx=b.link_idx,
      transform=b.transform,
      friction=b.friction,
      elasticity=b.elasticity,
  )


def get_face_norm(vert: jp.ndarray, face: jp.ndarray) -> jp.ndarray:
  """Calculates face normals given vertices and face indexes."""
  face_vert = jp.take(vert, face, axis=0)
  # use CCW winding order convention
  edge0 = face_vert[:, 0, :] - face_vert[:, 2, :]
  edge1 = face_vert[:, 0, :] - face_vert[:, 1, :]
  face_norm = jp.cross(edge0, edge1)
  face_norm = face_norm / jp.linalg.norm(face_norm, axis=1).reshape((-1, 1))
  return face_norm



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gotta go fast!  This trivial Env is meant for unit testing."""

import brax
from brax.envs import env
import jax.numpy as jnp


class Fast(env.Env):
  """Trains an agent to go fast."""

  def __init__(self, **kwargs):
    super().__init__(config='dt: .02', **kwargs)

  def reset(self, rng: jnp.ndarray) -> env.State:
    zero = jnp.zeros(1)
    qp = brax.QP(pos=zero, vel=zero, rot=zero, ang=zero)
    obs = jnp.zeros(2)
    reward, done = jnp.zeros(2)
    return env.State(qp, obs, reward, done)

  def step(self, state: env.State, action: jnp.ndarray) -> env.State:
    vel = state.qp.vel + (action > 0) * self.sys.config.dt
    pos = state.qp.pos + vel * self.sys.config.dt

    qp = state.qp.replace(pos=pos, vel=vel)
    obs = jnp.array([pos[0], vel[0]])
    reward = pos[0]

    return state.replace(qp=qp, obs=obs, reward=reward)

  @property
  def observation_size(self):
    return 2

  @property
  def action_size(self):
    return 1



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Input normalization utils."""

import functools
from typing import Any

import jax
import jax.numpy as jnp


def bcast_local_devices(value, local_devices_to_use=1):
  """Broadcasts an object to all local devices."""
  devices = jax.local_devices()[:local_devices_to_use]
  return jax.device_put_replicated(value, devices)


def synchronize_hosts():
  if jax.process_count() == 1:
    return
  # Make sure all processes stay up until the end of main.
  x = jnp.ones([jax.local_device_count()])
  x = jax.device_get(jax.pmap(lambda x: jax.lax.psum(x, 'i'), 'i')(x))
  assert x[0] == jax.device_count()


def _fingerprint(x: Any) -> float:
  sums = jax.tree_util.tree_map(jnp.sum, x)
  return jax.tree_util.tree_reduce(lambda x, y: x + y, sums)


def is_replicated(x: Any, axis_name: str) -> jnp.ndarray:
  """Returns whether x is replicated.

  Should be called inside a function pmapped along 'axis_name'
  Args:
    x: Object to check replication.
    axis_name: pmap axis_name.

  Returns:
    boolean whether x is replicated.
  """
  fp = _fingerprint(x)
  return jax.lax.pmin(
      fp, axis_name=axis_name) == jax.lax.pmax(
          fp, axis_name=axis_name)


def assert_is_replicated(x: Any, debug: Any = None):
  """Returns whether x is replicated.

  Should be called from a non-jitted code.
  Args:
    x: Object to check replication.
    debug: Debug message in case of failure.
  """
  f = functools.partial(is_replicated, axis_name='i')
  assert jax.pmap(f, axis_name='i')(x)[0], debug



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Calculations for generating contacts."""

from typing import Optional, Tuple

from brax.v2 import math
from brax.v2.base import (
    Box,
    Capsule,
    Contact,
    Geometry,
    Mesh,
    Plane,
    Sphere,
    System,
    Transform,
)
from brax.v2.geometry import math as geom_math
from brax.v2.geometry import mesh as geom_mesh
import jax
from jax import numpy as jp
from jax.tree_util import tree_map


def _combine(
    geom_a: Geometry, geom_b: Geometry
) -> Tuple[float, float, Tuple[int, int]]:
  # default is to take maximum, but can override
  friction = jp.maximum(geom_a.friction, geom_b.friction)
  elasticity = jp.maximum(geom_a.elasticity, geom_b.elasticity)
  link_idx = (
      geom_a.link_idx,
      geom_b.link_idx if geom_b.link_idx is not None else -1,
  )
  return friction, elasticity, link_idx


def _sphere_plane(sphere: Sphere, plane: Plane) -> Contact:
  """Calculates one contact between a sphere and a plane."""
  n = math.rotate(jp.array([0.0, 0.0, 1.0]), plane.transform.rot)
  t = jp.dot(sphere.transform.pos - plane.transform.pos, n)
  penetration = sphere.radius - t
  # halfway between contact points on sphere and on plane
  pos = sphere.transform.pos - n * (sphere.radius - 0.5 * penetration)
  c = Contact(pos, n, penetration, *_combine(sphere, plane))
  # add a batch dimension of size 1
  return tree_map(lambda x: jp.expand_dims(x, axis=0), c)


def _sphere_sphere(s_a: Sphere, s_b: Sphere) -> Contact:
  """Calculates one contact between two spheres."""
  n, dist = math.normalize(s_a.transform.pos - s_b.transform.pos)
  penetration = s_a.radius + s_b.radius - dist
  s_a_pos = s_a.transform.pos - n * s_a.radius
  s_b_pos = s_b.transform.pos + n * s_b.radius
  pos = (s_a_pos + s_b_pos) * 0.5
  c = Contact(pos, n, penetration, *_combine(s_a, s_b))
  # add a batch dimension of size 1
  return tree_map(lambda x: jp.expand_dims(x, axis=0), c)


def _sphere_capsule(sphere: Sphere, capsule: Capsule) -> Contact:
  """Calculates one contact between a sphere and a capsule."""
  segment = jp.array([0.0, 0.0, capsule.length * 0.5])
  segment = math.rotate(segment, capsule.transform.rot)
  pt = geom_math.closest_segment_point(
      capsule.transform.pos - segment,
      capsule.transform.pos + segment,
      sphere.transform.pos,
  )
  n, dist = math.normalize(sphere.transform.pos - pt)
  penetration = sphere.radius + capsule.radius - dist

  sphere_pos = sphere.transform.pos - n * sphere.radius
  cap_pos = pt + n * capsule.radius
  pos = (sphere_pos + cap_pos) * 0.5

  c = Contact(pos, n, penetration, *_combine(sphere, capsule))
  # add a batch dimension of size 1
  return tree_map(lambda x: jp.expand_dims(x, axis=0), c)


def _sphere_mesh(sphere: Sphere, mesh: Mesh) -> Contact:
  """Calculates contacts between a sphere and a mesh."""

  @jax.vmap
  def sphere_face(face):
    pt = mesh.transform.pos + jax.vmap(math.rotate, in_axes=[0, None])(
        face, mesh.transform.rot
    )
    p0, p1, p2 = pt[0, :], pt[1, :], pt[2, :]

    tri_p = geom_math.closest_triangle_point(p0, p1, p2, sphere.transform.pos)
    n = sphere.transform.pos - tri_p
    n, dist = math.normalize(n)
    penetration = sphere.radius - dist
    sph_p = sphere.transform.pos - n * sphere.radius
    pos = (tri_p + sph_p) * 0.5
    return Contact(pos, n, penetration, *_combine(sphere, mesh))

  return sphere_face(jp.take(mesh.vert, mesh.face, axis=0))


def _capsule_plane(capsule: Capsule, plane: Plane) -> Contact:
  """Calculates two contacts between a capsule and a plane."""
  segment = jp.array([0.0, 0.0, capsule.length * 0.5])
  segment = math.rotate(segment, capsule.transform.rot)

  results = []
  for off in [segment, -segment]:
    sphere = Sphere(
        link_idx=capsule.link_idx,
        transform=Transform.create(pos=capsule.transform.pos + off),
        friction=capsule.friction,
        elasticity=capsule.elasticity,
        radius=capsule.radius,
    )
    results.append(_sphere_plane(sphere, plane))

  return tree_map(lambda *x: jp.concatenate(x), *results)


def _capsule_capsule(cap_a: Capsule, cap_b: Capsule) -> Contact:
  """Calculates one contact between two capsules."""
  seg_a = jp.array([0.0, 0.0, cap_a.length * 0.5])
  seg_a = math.rotate(seg_a, cap_a.transform.rot)
  seg_b = jp.array([0.0, 0.0, cap_b.length * 0.5])
  seg_b = math.rotate(seg_b, cap_b.transform.rot)
  pt_a, pt_b = geom_math.closest_segment_to_segment_points(
      cap_a.transform.pos - seg_a,
      cap_a.transform.pos + seg_a,
      cap_b.transform.pos - seg_b,
      cap_b.transform.pos + seg_b,
  )
  n, dist = math.normalize(pt_a - pt_b)
  penetration = cap_a.radius + cap_b.radius - dist

  cap_a_pos = pt_a - n * cap_a.radius
  cap_b_pos = pt_b + n * cap_b.radius
  pos = (cap_a_pos + cap_b_pos) * 0.5

  c = Contact(pos, n, penetration, *_combine(cap_a, cap_b))
  # add a batch dimension of size 1
  return tree_map(lambda x: jp.expand_dims(x, axis=0), c)


def _capsule_mesh(capsule: Capsule, mesh: Mesh) -> Contact:
  """Calculates contacts between a capsule and a mesh."""

  @jax.vmap
  def capsule_face(face, face_norm):
    seg = jp.array([0.0, 0.0, capsule.length * 0.5])
    seg = math.rotate(seg, capsule.transform.rot)
    end_a, end_b = capsule.transform.pos - seg, capsule.transform.pos + seg

    tri_norm = math.rotate(face_norm, mesh.transform.rot)
    pt = mesh.transform.pos + jax.vmap(math.rotate, in_axes=[0, None])(
        face, mesh.transform.rot
    )
    p0, p1, p2 = pt[..., 0, :], pt[..., 1, :], pt[..., 2, :]

    seg_p, tri_p = geom_math.closest_segment_triangle_points(
        end_a, end_b, p0, p1, p2, tri_norm
    )
    n = seg_p - tri_p
    n, dist = math.normalize(n)
    penetration = capsule.radius - dist
    cap_p = seg_p - n * capsule.radius
    pos = (tri_p + cap_p) * 0.5
    return Contact(pos, n, penetration, *_combine(capsule, mesh))

  face_vert = jp.take(mesh.vert, mesh.face, axis=0)
  face_norm = geom_mesh.get_face_norm(mesh.vert, mesh.face)
  return capsule_face(face_vert, face_norm)


def _mesh_plane(mesh: Mesh, plane: Plane) -> Contact:
  """Calculates contacts between a mesh and a plane."""

  @jax.vmap
  def point_plane(vert):
    n = math.rotate(jp.array([0.0, 0.0, 1.0]), plane.transform.rot)
    pos = mesh.transform.pos + math.rotate(vert, mesh.transform.rot)
    penetration = jp.dot(plane.transform.pos - pos, n)
    return Contact(pos, n, penetration, *_combine(mesh, plane))

  return point_plane(mesh.vert)


_TYPE_FUN = {
    (Sphere, Plane): jax.vmap(_sphere_plane),
    (Sphere, Sphere): jax.vmap(_sphere_sphere),
    (Sphere, Capsule): jax.vmap(_sphere_capsule),
    (Sphere, Mesh): jax.vmap(_sphere_mesh),
    (Capsule, Plane): jax.vmap(_capsule_plane),
    (Capsule, Capsule): jax.vmap(_capsule_capsule),
    (Capsule, Mesh): jax.vmap(_capsule_mesh),
    (Mesh, Plane): jax.vmap(_mesh_plane),
}


def contact(sys: System, x: Transform) -> Optional[Contact]:
  """Calculates contacts in the system.

  Args:
    sys: system defining the kinematic tree and other properties
    x: link transforms in world frame

  Returns:
    Contact pytree, one row for each element in sys.contacts

  Raises:
    RuntimeError: if sys.contacts has an invalid type pair
  """

  contacts = []

  for geom_a, geom_b in sys.contacts:
    # convert Box geometry to Mesh
    geom_a = geom_mesh.box(geom_a) if isinstance(geom_a, Box) else geom_a
    geom_b = geom_mesh.box(geom_b) if isinstance(geom_b, Box) else geom_b

    # check both type signatures: a, b <> b, a
    fun = _TYPE_FUN.get((type(geom_a), type(geom_b)))
    if fun is None:
      fun = _TYPE_FUN.get((type(geom_b), type(geom_a)))
      if fun is None:
        raise RuntimeError(
            f'unrecognized collider pair: {type(geom_a)}, {type(geom_b)}'
        )
      geom_a, geom_b = geom_b, geom_a

    tx_a = x.take(geom_a.link_idx).vmap().do(geom_a.transform)
    geom_a = geom_a.replace(transform=tx_a)
    # OK for geom b to have no parent links, e.g. static terrain
    if geom_b.link_idx is not None:
      tx_b = x.take(geom_b.link_idx).vmap().do(geom_b.transform)
      geom_b = geom_b.replace(transform=tx_b)

    c = fun(geom_a, geom_b)  # type: ignore
    c = tree_map(jp.concatenate, c)
    contacts.append(c)

  if not contacts:
    return None

  return jax.tree_map(lambda *x: jp.concatenate(x), *contacts)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Integrator functions."""

from typing import Tuple
from brax.v2 import math
from brax.v2 import scan
from brax.v2.base import System
import jax
from jax import numpy as jp


def _integrate_q_axis(sys: System, q: jp.ndarray, qd: jp.ndarray) -> jp.ndarray:
  """Integrates next q for revolute/prismatic joints."""
  return q + qd * sys.dt


def _integrate_q_free(sys: System, q: jp.ndarray, qd: jp.ndarray) -> jp.ndarray:
  """Integrates next q for free joints."""
  rot, ang = q[3:7], qd[3:6]
  ang_norm = jp.linalg.norm(ang) + 1e-8
  axis = ang / ang_norm
  angle = sys.dt * ang_norm
  qrot = math.quat_rot_axis(axis, angle)
  rot = math.quat_mul(rot, qrot)
  rot = rot / jp.linalg.norm(rot)
  pos, vel = q[0:3], qd[0:3]
  pos += vel * sys.dt

  return jp.concatenate([pos, rot])


def integrate(
    sys: System, q: jp.ndarray, qd: jp.ndarray, qdd: jp.ndarray
) -> Tuple[jp.ndarray, jp.ndarray]:
  """Semi-implicit euler integration.

  Args:
    sys: system defining the kinematic tree and other properties
    q: joint angle vector. quaternions for free/spherical, scalar otherwise
    qd: joint velocity vector
    qdd: joint acceleration vector

  Returns:
    q: updated joint angle vector
    qd: updated joint velocity vector
  """

  qd += qdd * sys.dt

  def q_fn(typ, link, q, qd):
    q = q.reshape(link.transform.pos.shape[0], -1)
    qd = qd.reshape(link.transform.pos.shape[0], -1)
    fun = jax.vmap(
        {
            'f': _integrate_q_free,
            '1': _integrate_q_axis,
            '2': _integrate_q_axis,
            '3': _integrate_q_axis,
        }[typ],
        in_axes=(None, 0, 0),
    )
    q_s = fun(sys, q, qd).reshape(-1)

    return q_s

  q = scan.link_types(sys, q_fn, 'lqd', 'q', sys.link, q, qd)

  return q, qd



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for math."""

from absl.testing import absltest
from brax import jumpy as jp
from brax import math


class RotateTest(absltest.TestCase):
  """Tests math.rotate."""

  def test_rotate(self):
    vec = jp.array([0, 0, 1])
    quat = math.euler_to_quat(jp.array([0, 45, 0]))
    self.assertSequenceAlmostEqual(math.rotate(vec, quat),
                                   [1. / jp.sqrt(2), 0, 1. / jp.sqrt(2)])

  def test_rotate_identity(self):
    vec = jp.ones(3)
    quat = jp.array([1, 0, 0, 0])
    self.assertSequenceAlmostEqual(math.rotate(vec, quat), vec)

  def test_rotate_bad_input_shape(self):
    vec = jp.ones((3, 3))
    quat = jp.array([1, 0, 0, 0])
    with self.assertRaises(AssertionError):
      math.rotate(vec, quat)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pytree decorator for classes."""
import inspect
import jax

# Rolling our own instead of using flax.struct.dataclass because we:
# - would like our own __init__ fns / super() without too much hassle
# - want child classes to inherit from parent classes that have pytree=None
# - aren't worried about immutability / don't like dataclasses.replace()


def register(cls):
  """Registers a class to become a pytree node.

  Treats any class fields as pytree nodes unless they are added to a
  '__pytree_ignore__' class attribute.

  Args:
    cls: the class to register.

  Returns:
    the input class unchanged.
  """

  def tree_flatten(obj):
    pytree_data = []
    pytree_fields = []
    static_data = {}
    static_fields = set()
    for c in inspect.getmro(cls):
      if hasattr(c, '__pytree_ignore__'):
        static_fields.update(cls.__pytree_ignore__)
    for k, v in vars(obj).items():
      if k in static_fields:
        static_data[k] = v
      else:
        pytree_fields.append(k)
        pytree_data.append(v)
    return (pytree_data, (pytree_fields, static_data))

  def tree_unflatten(aux_data, children):
    obj = cls.__new__(cls)
    pytree_fields, static_fields = aux_data
    for k, v in zip(pytree_fields, children):
      setattr(obj, k, v)
    for k, v in static_fields.items():
      setattr(obj, k, v)
    return obj

  jax.tree_util.register_pytree_node(cls, tree_flatten, tree_unflatten)

  return cls



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An inverted pendulum environment."""

import brax
from brax import jumpy as jp
from brax.envs import env


class InvertedPendulum(env.Env):



  """
  ### Description

  This environment is the cartpole environment based on the work done by Barto,
  Sutton, and Anderson in
  ["Neuronlike adaptive elements that can solve difficult learning control problems"](https://ieeexplore.ieee.org/document/6313077).

  This environment involves a cart that can moved linearly, with a pole fixed on
  it at one end and having another end free. The cart can be pushed left or
  right, and the goal is to balance the pole on the top of the cart by applying
  forces on the cart.

  ### Action Space

  The agent take a 1-element vector for actions. The action space is a
  continuous `(action)` in `[-3, 3]`, where `action` represents the numerical
  force applied to the cart (with magnitude representing the amount of force and
  sign representing the direction)

  | Num | Action                    | Control Min | Control Max | Name (in corresponding config) | Joint | Unit      |
  |-----|---------------------------|-------------|-------------|--------------------------------|-------|-----------|
  | 0   | Force applied on the cart | -1          | 1           | thruster                       | slide | Force (N) |

  ### Observation Space

  The state space consists of positional values of different body parts of the
  pendulum system, followed by the velocities of those individual parts (their
  derivatives) with all the positions ordered before all the velocities.

  The observation is a `ndarray` with shape `(4,)` where the elements correspond
  to the following:

  | Num | Observation                                   | Min  | Max | Name (in corresponding config) | Joint | Unit                     |
  |-----|-----------------------------------------------|------|-----|--------------------------------|-------|--------------------------|
  | 0   | position of the cart along the linear surface | -Inf | Inf | thruster                       | slide | position (m)             |
  | 1   | vertical angle of the pole on the cart        | -Inf | Inf | hinge                          | hinge | angle (rad)              |
  | 2   | linear velocity of the cart                   | -Inf | Inf | thruster                       | slide | velocity (m/s)           |
  | 3   | angular velocity of the pole on the cart      | -Inf | Inf | hinge                          | hinge | angular velocity (rad/s) |


  ### Rewards

  The goal is to make the inverted pendulum stand upright (within a certain
  angle limit) as long as possible - as such a reward of +1 is awarded for each
  timestep that the pole is upright.

  ### Starting State
  All observations start in state (0.0, 0.0, 0.0, 0.0) with a uniform noise in
  the range of [-0.01, 0.01] added to the values for stochasticity.

  ### Episode Termination

  The episode terminates when any of the following happens:

  1. The episode duration reaches 1000 timesteps.
  2. The absolute value of the vertical angle between the pole and the cart is
  greater than 0.2 radians.

  ### Arguments

  No additional arguments are currently supported (in v2 and lower), but
  modifications can be made to the XML file in the assets folder (or by changing
  the path to a modified XML file in another folder).

  ```
  env = gym.make('InvertedPendulum-v2')
  ```

  There is no v3 for InvertedPendulum, unlike the robot environments where a v3
  and beyond take gym.make kwargs such as ctrl_cost_weight, reset_noise_scale
  etc.

  There is a v4 version that uses the mujoco-bindings

  ```
  env = gym.make('InvertedPendulum-v4')
  ```

  And a v5 version that uses Brax:

  ```
  env = gym.make('InvertedPendulum-v5')
  ```

  ### Version History

  * v5: ported to Brax.
  * v4: all mujoco environments now use the mujoco bindings in mujoco>=2.1.3
  * v3: support for gym.make kwargs such as ctrl_cost_weight, reset_noise_scale
    etc. rgb rendering comes from tracking camera (so agent does not run away
    from screen)
  * v2: All continuous control environments now use mujoco_py >= 1.50
  * v1: max_time_steps raised to 1000 for robot based tasks (including inverted
    pendulum)
  * v0: Initial versions release (1.0.0)
  """


  def __init__(self, legacy_spring=False, **kwargs):
    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
    super().__init__(config=config, **kwargs)

  def reset(self, rng: jp.ndarray) -> env.State:
    """Resets the environment to an initial state."""
    rng, rng1, rng2 = jp.random_split(rng, 3)

    qpos = self.sys.default_angle() + self._noise(rng1)
    qvel = self._noise(rng2)

    qp = self.sys.default_qp(joint_angle=qpos, joint_velocity=qvel)
    obs = self._get_obs(qp, self.sys.info(qp))
    reward, done = jp.zeros(2)
    metrics = {}

    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    """Run one timestep of the environment's dynamics."""
    qp, info = self.sys.step(state.qp, action)
    obs = self._get_obs(qp, info)
    reward = 1.0
    done = jp.where(jp.abs(obs[1]) > .2, 1.0, 0.0)

    return state.replace(qp=qp, obs=obs, reward=reward, done=done)

  @property
  def action_size(self):
    return 1

  def _get_obs(self, qp: brax.QP, info: brax.Info) -> jp.ndarray:
    """Observe cartpole body position and velocities."""
    joint_angle, joint_vel = self.sys.joints[0].angle_vel(qp)

    # [cart pos, joint angle, cart vel, joint vel]
    obs = [qp.pos[0, :1], joint_angle, qp.vel[0, :1], joint_vel]

    return jp.concatenate(obs)

  def _noise(self, rng):
    return jp.random_uniform(rng, (self.sys.num_joint_dof,), -0.01, 0.01)


_SYSTEM_CONFIG = """
  bodies {
    name: "cart"
    colliders {
      rotation {
      x: 90
      z: 90
      }
      capsule {
        radius: 0.1
        length: 0.4
      }
    }
    frozen { position { x:0 y:1 z:1 } rotation { x:1 y:1 z:1 } }
    mass: 10.471975
  }
  bodies {
    name: "pole"
    colliders {
      capsule {
        radius: 0.049
        length: 0.69800085
      }
    }
    frozen { position { x: 0 y: 1 z: 0 } rotation { x: 1 y: 0 z: 1 } }
    mass: 5.0185914
  }
  joints {
    name: "hinge"
    parent: "cart"
    child: "pole"
    child_offset { z: -.3 }
    rotation {
      z: 90.0
    }
    angle_limit { min: -360 max: 360 }
  }
  forces {
    name: "thruster"
    body: "cart"
    strength: 100.0
    thruster {}
  }
  collide_include {}
  gravity {
    z: -9.81
  }
  dt: 0.04
  substeps: 8
  dynamics_mode: "pbd"
  """


_SYSTEM_CONFIG_SPRING = """
  bodies {
    name: "cart"
    colliders {
      rotation {
      x: 90
      z: 90
      }
      capsule {
        radius: 0.1
        length: 0.4
      }
    }
    frozen { position { x:0 y:1 z:1 } rotation { x:1 y:1 z:1 } }
    mass: 10.471975
  }
  bodies {
    name: "pole"
    colliders {
      capsule {
        radius: 0.049
        length: 0.69800085
      }
    }
    frozen { position { x: 0 y: 1 z: 0 } rotation { x: 1 y: 0 z: 1 } }
    mass: 5.0185914
  }
  joints {
    name: "hinge"
    stiffness: 10000.0
    parent: "cart"
    child: "pole"
    child_offset { z: -.3 }
    rotation {
      z: 90.0
    }
    limit_strength: 0.0
    angle_limit { min: 0.0 max: 0.0 }
  }
  forces {
    name: "thruster"
    body: "cart"
    strength: 100.0
    thruster {}
  }
  collide_include {}
  gravity {
    z: -9.81
  }
  dt: 0.04
  substeps: 8
  dynamics_mode: "legacy_spring"
  """



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simulator utilities."""

import collections
import functools
from typing import List, Tuple, Any
import brax
import jax
from jax import numpy as jnp
from brax import jumpy as jp

lim_to_dof = {0: 1, 1: 1, 2: 2, 3: 3}


@functools.partial(jax.vmap, in_axes=[0, 0, None, None, None])
def transform_qp(qp, mask: jnp.ndarray, rot: jnp.ndarray, rot_vec: jnp.ndarray,
                 offset_vec: jnp.ndarray):
  """Rotates a qp by some rot around some ref_vec and translates it.

  Args:
    qp: QPs to be rotated
    mask: whether to transform this qp or not
    rot: Quaternion to rotate by
    rot_vec: point around which to rotate.
    offset_vec: relative displacement vector to translate qp by

  Returns:
    transformed QP
  """
  relative_pos = qp.pos - rot_vec
  new_pos = brax.math.rotate(relative_pos, rot) + rot_vec + offset_vec
  new_rot = brax.math.quat_mul(rot, qp.rot)
  return brax.physics.base.QP(
      pos=jnp.where(mask, new_pos, qp.pos),
      vel=qp.vel,
      ang=qp.ang,
      rot=jnp.where(mask, new_rot, qp.rot))


def get_names(config, datatype: str = 'body'):
  objs = {
      'body': config.bodies,
      'joint': config.joints,
      'actuator': config.actuators,
      'force': config.forces,
  }[datatype]
  return [b.name for b in objs]


def _legacy_angle_vel(joint, qp) -> Tuple[Any, Any]:
  """Returns joint angle and velocity using the legacy format.

  Args:
    joint: Joint to operate over
    qp: State data for system

  Returns:
    angle: n-tuple of joint angles where n = # DoF of the joint
    vel: n-tuple of joint velocities where n = # DoF of the joint
  """

  @jax.vmap
  def op(joint, qp_p, qp_c):
    axes, angles = joint.axis_angle(qp_p, qp_c)
    vels = tuple([jnp.dot(qp_p.ang - qp_c.ang, axis) for axis in axes])
    return angles, vels

  qp_p = jp.take(qp, joint.body_p.idx)
  qp_c = jp.take(qp, joint.body_c.idx)
  angles, vels = op(joint, qp_p, qp_c)

  return angles, vels


def get_joint_value(sys, qp, info: collections.OrderedDict):
  """Get joint values."""
  values = collections.OrderedDict()
  angles_vels = {j.dof: _legacy_angle_vel(j, qp) for j in sys.joints}
  for k, v in info.items():
    for i, type_ in zip((0, 1), ('pos', 'vel')):
      vvv = jnp.array([vv[v['index']] for vv in angles_vels[v['dof']][i]])
      values[f'joint_{type_}:{k}'] = vvv
  return values


def _check_active_dofs(joint):
  active_dof = 0
  for l in joint.angle_limit:
    if not (l.min == 0 and l.max == 0):
      active_dof += 1
  return active_dof


def names2indices(config, names: List[str], datatype: str = 'body'):
  """Convert name string to indices for indexing simulator states."""

  if isinstance(names, str):
    names = [names]

  indices = {}
  info = {}

  objs = {
      'body': config.bodies,
      'joint': config.joints,
      'actuator': config.actuators,
      'force': config.forces,
  }[datatype]
  joint_counters = [0, 0, 0]
  actuator_counter = 0
  force_counter = 0
  for i, b in enumerate(objs):
    if datatype == 'joint':
      if config.dynamics_mode == 'pbd':
        dof = _check_active_dofs(b)
      else:
        dof = lim_to_dof[len(b.angle_limit)]
    elif datatype == 'actuator':
      joint = [j for j in config.joints if j.name == b.joint][0]
      if config.dynamics_mode == 'pbd':
        dof = _check_active_dofs(joint)
      else:
        dof = lim_to_dof[len(joint.angle_limit)]
    elif datatype == 'force':
      dof = 3
    if b.name in names:
      indices[b.name] = i
      if datatype in ('joint',):
        info[b.name] = dict(dof=dof, index=joint_counters[dof - 1])
      if datatype in ('actuator',):
        info[b.name] = tuple(range(actuator_counter, actuator_counter + dof))
      if datatype in ('force',):
        info[b.name] = tuple(range(force_counter, force_counter + dof))
    if datatype in ('joint',):
      joint_counters[dof - 1] += 1
    if datatype in ('actuator',):
      actuator_counter += dof
    if datatype in ('force',):
      force_counter += dof

  indices = [indices[n] for n in names]
  mask = jnp.array([b.name in names for b in objs])

  if datatype in ('actuator', 'force', 'joint'):
    info = collections.OrderedDict([(k, info[k]) for k in names])

  return indices, info, mask



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrappers for Brax and Gym env."""

import copy
from typing import ClassVar, Dict, Optional

from brax import jumpy as jp
from brax.envs import env as brax_env
import dm_env
from dm_env import specs
import flax
import gym
from gym import spaces
from gym.vector import utils
import jax


def wrap_for_training(env: brax_env.Env,
                      episode_length: int = 1000,
                      action_repeat: int = 1) -> brax_env.Wrapper:
  """Common wrapper pattern for all training agents.

  Args:
    env: environment to be wrapped
    episode_length: length of episode
    action_repeat: how many repeated actions to take per step

  Returns:
    An environment that is wrapped with Episode and AutoReset wrappers.  If the
    environment did not already have batch dimensions, it is additional Vmap
    wrapped.
  """
  env = EpisodeWrapper(env, episode_length, action_repeat)
  batched = False
  if hasattr(env, 'custom_tree_in_axes'):
    batch_indices, _ = jax.tree_util.tree_flatten(env.custom_tree_in_axes)
    if 0 in batch_indices:
      batched = True
  if not batched:
    env = VmapWrapper(env)
  env = AutoResetWrapper(env)
  return env


class VectorWrapper(brax_env.Wrapper):
  """DEPRECATED Vectorizes Brax env. Use VmapWrapper instead."""

  def __init__(self, env: brax_env.Env, batch_size: int):
    super().__init__(env)
    self.batch_size = batch_size

  def reset(self, rng: jp.ndarray) -> brax_env.State:
    rng = jp.random_split(rng, self.batch_size)
    return jp.vmap(self.env.reset)(rng)

  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:
    return jp.vmap(self.env.step)(state, action)


class VmapWrapper(brax_env.Wrapper):
  """Vectorizes Brax env."""

  def reset(self, rng: jp.ndarray) -> brax_env.State:
    return jp.vmap(self.env.reset)(rng)

  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:
    return jp.vmap(self.env.step)(state, action)


class EpisodeWrapper(brax_env.Wrapper):
  """Maintains episode step count and sets done at episode end."""

  def __init__(self, env: brax_env.Env, episode_length: int,
               action_repeat: int):
    super().__init__(env)
    # For proper video speed.
    # TODO: Fix dt book keeping with action repeats so there isn't
    # async between sys and env steps.
    if hasattr(env, 'sys'):
      self.sys = copy.deepcopy(env.sys)
      self.sys.config.dt *= action_repeat
      self.sys.config.substeps *= action_repeat
    self.episode_length = episode_length
    self.action_repeat = action_repeat

  def reset(self, rng: jp.ndarray) -> brax_env.State:
    state = self.env.reset(rng)
    state.info['steps'] = jp.zeros(rng.shape[:-1])
    state.info['truncation'] = jp.zeros(rng.shape[:-1])
    return state

  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:
    def f(state, _):
      nstate = self.env.step(state, action)
      return nstate, nstate.reward

    state, rewards = jp.scan(f, state, (), self.action_repeat)
    state = state.replace(reward=jp.sum(rewards, axis=0))
    steps = state.info['steps'] + self.action_repeat
    one = jp.ones_like(state.done)
    zero = jp.zeros_like(state.done)
    episode_length = jp.array(self.episode_length, dtype=jp.int32)
    done = jp.where(steps >= episode_length, one, state.done)
    state.info['truncation'] = jp.where(steps >= episode_length,
                                        1 - state.done, zero)
    state.info['steps'] = steps
    return state.replace(done=done)


class AutoResetWrapper(brax_env.Wrapper):
  """Automatically resets Brax envs that are done."""

  def reset(self, rng: jp.ndarray) -> brax_env.State:
    state = self.env.reset(rng)
    state.info['first_qp'] = state.qp
    state.info['first_obs'] = state.obs
    return state

  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:
    if 'steps' in state.info:
      steps = state.info['steps']
      steps = jp.where(state.done, jp.zeros_like(steps), steps)
      state.info.update(steps=steps)
    state = state.replace(done=jp.zeros_like(state.done))
    state = self.env.step(state, action)

    def where_done(x, y):
      done = state.done
      if done.shape:
        done = jp.reshape(done, [x.shape[0]] + [1] * (len(x.shape) - 1))  # type: ignore
      return jp.where(done, x, y)

    qp = jp.tree_map(where_done, state.info['first_qp'], state.qp)
    obs = where_done(state.info['first_obs'], state.obs)
    return state.replace(qp=qp, obs=obs)


@flax.struct.dataclass
class EvalMetrics:
  """Dataclass holding evaluation metrics for Brax.

    Args:
        episode_metrics: Aggregated episode metrics since the beginning of the
          episode.
        active_episodes: Boolean vector tracking which episodes are not done
          yet.
        episode_steps: Integer vector tracking the number of steps in
          the episode.
  """
  episode_metrics: Dict[str, jp.ndarray]
  active_episodes: jp.ndarray
  episode_steps: jp.ndarray


class EvalWrapper(brax_env.Wrapper):
  """Brax env with eval metrics."""

  def reset(self, rng: jp.ndarray) -> brax_env.State:
    reset_state = self.env.reset(rng)
    reset_state.metrics['reward'] = reset_state.reward
    eval_metrics = EvalMetrics(
        episode_metrics=jax.tree_util.tree_map(jp.zeros_like,
                                               reset_state.metrics),
        active_episodes=jp.ones_like(reset_state.reward),
        episode_steps=jp.zeros_like(reset_state.reward))
    reset_state.info['eval_metrics'] = eval_metrics
    return reset_state

  def step(self, state: brax_env.State, action: jp.ndarray) -> brax_env.State:
    state_metrics = state.info['eval_metrics']
    if not isinstance(state_metrics, EvalMetrics):
      raise ValueError(
          f'Incorrect type for state_metrics: {type(state_metrics)}')
    del state.info['eval_metrics']
    nstate = self.env.step(state, action)
    nstate.metrics['reward'] = nstate.reward
    episode_steps = jp.where(state_metrics.active_episodes,
                             nstate.info['steps'], state_metrics.episode_steps)
    episode_metrics = jax.tree_util.tree_map(
        lambda a, b: a + b * state_metrics.active_episodes,
        state_metrics.episode_metrics, nstate.metrics)
    active_episodes = state_metrics.active_episodes * (1 - nstate.done)

    eval_metrics = EvalMetrics(
        episode_metrics=episode_metrics,
        active_episodes=active_episodes,
        episode_steps=episode_steps)
    nstate.info['eval_metrics'] = eval_metrics
    return nstate


class GymWrapper(gym.Env):
  """A wrapper that converts Brax Env to one that follows Gym API."""

  # Flag that prevents `gym.register` from misinterpreting the `_step` and
  # `_reset` as signs of a deprecated gym Env API.
  _gym_disable_underscore_compat: ClassVar[bool] = True

  def __init__(self,
               env: brax_env.Env,
               seed: int = 0,
               backend: Optional[str] = None):
    self._env = env
    self.metadata = {
        'render.modes': ['human', 'rgb_array'],
        'video.frames_per_second': 1 / self._env.sys.config.dt
    }
    self.seed(seed)
    self.backend = backend
    self._state = None

    obs_high = jp.inf * jp.ones(self._env.observation_size, dtype='float32')
    self.observation_space = spaces.Box(-obs_high, obs_high, dtype='float32')

    action_high = jp.ones(self._env.action_size, dtype='float32')
    self.action_space = spaces.Box(-action_high, action_high, dtype='float32')

    def reset(key):
      key1, key2 = jp.random_split(key)
      state = self._env.reset(key2)
      return state, state.obs, key1

    self._reset = jax.jit(reset, backend=self.backend)

    def step(state, action):
      state = self._env.step(state, action)
      info = {**state.metrics, **state.info}
      return state, state.obs, state.reward, state.done, info

    self._step = jax.jit(step, backend=self.backend)

  def reset(self):
    self._state, obs, self._key = self._reset(self._key)
    # We return device arrays for pytorch users.
    return obs

  def step(self, action):
    self._state, obs, reward, done, info = self._step(self._state, action)
    # We return device arrays for pytorch users.
    return obs, reward, done, info

  def seed(self, seed: int = 0):
    self._key = jax.random.PRNGKey(seed)

  def render(self, mode='human'):
    # pylint:disable=g-import-not-at-top
    from brax.io import image
    if mode == 'rgb_array':
      sys, qp = self._env.sys, self._state.qp
      return image.render_array(sys, qp, 256, 256)
    else:
      return super().render(mode=mode)  # just raise an exception


class VectorGymWrapper(gym.vector.VectorEnv):
  """A wrapper that converts batched Brax Env to one that follows Gym VectorEnv API."""

  # Flag that prevents `gym.register` from misinterpreting the `_step` and
  # `_reset` as signs of a deprecated gym Env API.
  _gym_disable_underscore_compat: ClassVar[bool] = True

  def __init__(self,
               env: brax_env.Env,
               seed: int = 0,
               backend: Optional[str] = None):
    self._env = env
    self.metadata = {
        'render.modes': ['human', 'rgb_array'],
        'video.frames_per_second': 1 / self._env.sys.config.dt
    }
    if not hasattr(self._env, 'batch_size'):
      raise ValueError('underlying env must be batched')

    self.num_envs = self._env.batch_size
    self.seed(seed)
    self.backend = backend
    self._state = None

    obs_high = jp.inf * jp.ones(self._env.observation_size, dtype='float32')
    self.single_observation_space = spaces.Box(
        -obs_high, obs_high, dtype='float32')
    self.observation_space = utils.batch_space(self.single_observation_space,
                                               self.num_envs)

    action_high = jp.ones(self._env.action_size, dtype='float32')
    self.single_action_space = spaces.Box(
        -action_high, action_high, dtype='float32')
    self.action_space = utils.batch_space(self.single_action_space,
                                          self.num_envs)

    def reset(key):
      key1, key2 = jp.random_split(key)
      state = self._env.reset(key2)
      return state, state.obs, key1

    self._reset = jax.jit(reset, backend=self.backend)

    def step(state, action):
      state = self._env.step(state, action)
      info = {**state.metrics, **state.info}
      return state, state.obs, state.reward, state.done, info

    self._step = jax.jit(step, backend=self.backend)

  def reset(self):
    self._state, obs, self._key = self._reset(self._key)
    return obs

  def step(self, action):
    self._state, obs, reward, done, info = self._step(self._state, action)
    return obs, reward, done, info

  def seed(self, seed: int = 0):
    self._key = jax.random.PRNGKey(seed)

  def render(self, mode='human'):
    # pylint:disable=g-import-not-at-top
    from brax.io import image
    if mode == 'rgb_array':
      sys = self._env.sys
      qp = jp.take(self._state.qp, 0)
      return image.render_array(sys, qp, 256, 256)
    else:
      return super().render(mode=mode)  # just raise an exception


class DmEnvWrapper(dm_env.Environment):
  """A wrapper that converts Brax Env to one that follows Dm Env API."""

  def __init__(self,
               env: brax_env.Env,
               seed: int = 0,
               backend: Optional[str] = None):
    self._env = env
    self.seed(seed)
    self.backend = backend
    self._state = None

    if hasattr(self._env, 'observation_spec'):
      self._observation_spec = self._env.observation_spec()
    else:
      obs_high = jp.inf * jp.ones(self._env.observation_size, dtype='float32')
      self._observation_spec = specs.BoundedArray((self._env.observation_size,),
                                                  minimum=-obs_high,
                                                  maximum=obs_high,
                                                  dtype='float32',
                                                  name='observation')

    if hasattr(self._env, 'action_spec'):
      self._action_spec = self._env.action_spec()
    else:
      action_high = jp.ones(self._env.action_size, dtype='float32')
      self._action_spec = specs.BoundedArray((self._env.action_size,),
                                             minimum=-action_high,
                                             maximum=action_high,
                                             dtype='float32',
                                             name='action')

    self._reward_spec = specs.Array(shape=(), dtype='float32', name='reward')
    self._discount_spec = specs.BoundedArray(
        shape=(), dtype='float32', minimum=0., maximum=1., name='discount')
    if hasattr(self._env, 'discount_spec'):
      self._discount_spec = self._env.discount_spec()

    def reset(key):
      key1, key2 = jp.random_split(key)
      state = self._env.reset(key2)
      return state, state.obs, key1

    self._reset = jax.jit(reset, backend=self.backend)

    def step(state, action):
      state = self._env.step(state, action)
      info = {**state.metrics, **state.info}
      return state, state.obs, state.reward, state.done, info

    self._step = jax.jit(step, backend=self.backend)

  def reset(self):
    self._state, obs, self._key = self._reset(self._key)
    return dm_env.TimeStep(
        step_type=dm_env.StepType.FIRST,
        reward=None,
        discount=jp.float32(1.),
        observation=obs)

  def step(self, action):
    self._state, obs, reward, done, info = self._step(self._state, action)
    del info
    return dm_env.TimeStep(
        step_type=dm_env.StepType.MID if not done else dm_env.StepType.LAST,
        reward=reward,
        discount=jp.float32(1.),
        observation=obs)

  def seed(self, seed: int = 0):
    self._key = jax.random.PRNGKey(seed)

  def observation_spec(self):
    return self._observation_spec

  def action_spec(self):
    return self._action_spec

  def reward_spec(self):
    return self._reward_spec

  def discount_spec(self):
    return self._discount_spec

  def render(self):
    # pylint:disable=g-import-not-at-top
    from brax.io import image
    sys, qp = self._env.sys, self._state.qp
    return image.render_array(sys, qp, 256, 256)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluators."""
import functools
import os
from typing import Dict, Tuple, Any
from brax.experimental.braxlines.training import utils as training_utils
from brax.experimental.braxlines.training.env import wrap
from brax.io import html
import jax
from jax import numpy as jnp


def init_run_eval(env_fn,
                  action_repeat: int,
                  batch_size: int,
                  episode_length: int = 1000,
                  seed: int = 0):
  """Initialize run_eval."""
  key_eval = jax.random.PRNGKey(seed=seed)
  core_eval_env = env_fn(
      action_repeat=action_repeat,
      batch_size=batch_size,
      episode_length=episode_length)
  eval_first_state, eval_step_fn = wrap(core_eval_env, key_eval)
  return core_eval_env, eval_first_state, eval_step_fn, key_eval


def jit_run_eval(env_fn,
                 inference_fn,
                 action_repeat: int,
                 batch_size: int,
                 episode_length: int = 1000,
                 seed: int = 0):
  """JIT run_eval."""
  core_eval_env, eval_first_state, eval_step_fn, key_eval = init_run_eval(
      env_fn=env_fn,
      seed=seed,
      action_repeat=action_repeat,
      batch_size=batch_size,
      episode_length=episode_length)
  jit_inference_fn = jax.jit(inference_fn)

  def do_one_step_eval(carry, unused_target_t):
    state, params, key = carry
    key, key_sample = jax.random.split(key)
    actions = jit_inference_fn(params, state.core.obs, key_sample)
    nstate = eval_step_fn(state, actions, params['normalizer'], params['extra'])
    return (nstate, params, key), nstate.core

  @jax.jit
  def jit_run_eval_fn(state, key, params):
    (final_state, _, key), states = jax.lax.scan(
        do_one_step_eval, (state, params, key), (),
        length=episode_length // action_repeat)
    return final_state, key, states

  return core_eval_env, eval_first_state, key_eval, jit_run_eval_fn


def rollout_env(
    params: Dict[str, Dict[str, jnp.ndarray]],
    env_fn,
    inference_fn,
    batch_size: int = 0,
    seed: int = 0,
    reset_args: Tuple[Any] = (),
    step_args: Tuple[Any] = (),
    step_fn_name: str = 'step',
):
  """Visualize environment."""
  rng = jax.random.PRNGKey(seed=seed)
  rng, reset_key = jax.random.split(rng)
  env = env_fn(batch_size=batch_size)
  inference_fn = inference_fn or functools.partial(
      training_utils.zero_fn, action_size=env.action_size)
  jit_env_reset = jax.jit(env.reset)
  jit_env_step = jax.jit(getattr(env, step_fn_name))
  jit_inference_fn = jax.jit(inference_fn)
  states = []
  state = jit_env_reset(reset_key, *reset_args)
  while not jnp.all(state.done):
    states.append(state)
    tmp_key, rng = jax.random.split(rng)
    act = jit_inference_fn(params, state.obs, tmp_key)
    state = jit_env_step(state, act, *step_args)
  states.append(state)
  return env, states


def visualize_env(batch_size: int = 0,
                  output_path: str = None,
                  output_name: str = 'video',
                  verbose: bool = False,
                  **kwargs):
  """Visualize env."""
  env, states = rollout_env(batch_size=batch_size, **kwargs)
  if verbose:
    print(f'Collected {max(1, batch_size)} trajs of T={len(states)}')

  if output_path:
    output_name = os.path.splitext(output_name)[0]
    if batch_size:

      for i in range(batch_size):
        filename = f'{output_path}/{output_name}_eps{i:02}.html'
        html.save_html(
            filename,
            env.sys, [
                jax.tree_util.tree_map(functools.partial(jnp.take, indices=i), state.qp)
                for state in states
            ],
            make_dir=True)
        if verbose:
          print(f'Saved {filename}')
    else:
      filename = f'{output_path}/{output_name}.html'
      html.save_html(
          filename, env.sys, [state.qp for state in states], make_dir=True)
      if verbose:
        print(f'Saved {filename}')

  return env, states



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains an claw hand to grasp an object and move it to targets."""

from typing import Tuple

import brax
from brax import jumpy as jp
from brax import math
from brax.envs import env


class Grasp(env.Env):
  """Grasp trains an agent to pick up an object.

  Grasp observes three bodies: 'Hand', 'Object', and 'Target'.
  When Object reaches Target, the agent is rewarded.
  """

  def __init__(self, legacy_spring=False, **kwargs):
    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
    super().__init__(config=config, **kwargs)
    self.object_idx = self.sys.body.index['Object']
    self.target_idx = self.sys.body.index['Target']
    self.hand_idx = self.sys.body.index['HandThumbProximal']
    self.palm_idx = self.sys.body.index['HandPalm']
    self.target_radius = 1.1
    self.target_distance = 10.
    self.target_height = 8.

    # map the [-1, 1] action space into a valid angle for the actuators
    limits = []
    for j in self.sys.config.joints:
      for l in j.angle_limit:
        limits.append((l.min, l.max))
    self._min_act = jp.array([l[0] for l in limits])
    self._range_act = jp.array([l[1] - l[0] for l in limits])

    # add limits for the translational motion of the hand base
    self._min_act = jp.concatenate([self._min_act, jp.array([-10, -10, 3.5])])
    self._range_act = jp.concatenate([self._range_act, jp.array([20, 20, 10])])

  def reset(self, rng: jp.ndarray) -> env.State:
    qp = self.sys.default_qp()
    # rng, target = self._random_target(rng)
    # pos = qp.pos.at[self.target_idx].set(target)
    # qp = dataclasses.replace(qp, pos=pos)
    info = self.sys.info(qp)
    obs = self._get_obs(qp, info)
    reward, done, zero = jp.zeros(3)
    metrics = {
        'hits': zero,
        'touchingObject': zero,
        'movingToObject': zero,
        'movingObjectToTarget': zero,
        'closeToObject': zero
    }
    info = {'rng': rng}
    return env.State(qp, obs, reward, done, metrics, info)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    # actuate the palm
    action = self._min_act + self._range_act * ((action + 1) / 2.)
    target_pos = action[-3:]
    palm_pos = state.qp.pos[self.palm_idx]
    norm = jp.norm(target_pos - palm_pos)
    # make sure hand doesn't move too fast
    scale = jp.where(norm > 2.0, 2. / norm, 1.0)
    palm_pos = palm_pos + scale * (target_pos - palm_pos) * .15
    pos = state.qp.pos
    pos = jp.index_update(pos, self.palm_idx, palm_pos)
    qp = state.qp.replace(pos=pos)

    # do the rest of the physics update
    qp, info = self.sys.step(qp, action)
    obs = self._get_obs(qp, info)

    # small reward for moving in right direction
    object_pos = qp.pos[self.object_idx]
    hand_pos = qp.pos[self.palm_idx]
    hand_vel = qp.vel[self.hand_idx]
    object_rel = object_pos - hand_pos
    object_dist = jp.norm(object_rel)
    planar_object_dist = jp.norm(object_rel * jp.array([1., 1., 0.]))
    object_dir = object_rel / (1e-6 + object_dist.reshape(-1))

    moving_to_object = .1 * self.sys.config.dt * jp.dot(hand_vel, object_dir)
    close_to_object = .1 * self.sys.config.dt * 1. / (1. + planar_object_dist)

    # large reward for object moving to target
    target_pos = qp.pos[self.target_idx]
    object_vel = qp.vel[self.object_idx]
    target_rel = target_pos - object_pos

    target_dist = jp.norm(target_rel)
    target_dir = target_rel / (1e-6 + target_dist)
    moving_to_target = 1.5 * self.sys.config.dt * jp.dot(
        object_vel, target_dir)

    # small reward for touching object
    contact_mag = jp.sum(jp.square(info.contact.vel), axis=-1)
    contacts = jp.where(contact_mag > 0.00001, 1, 0)
    touching_object = 0.2 * self.sys.config.dt * (
        contacts[3] + contacts[9] + contacts[12] + contacts[15])

    # big reward for reaching target
    target_hit = jp.where(target_dist < self.target_radius, 1.0, 0.0)

    reward = moving_to_object + close_to_object + touching_object + 5. * target_hit + moving_to_target

    state.metrics.update(
        hits=target_hit,
        touchingObject=touching_object,
        movingToObject=moving_to_object,
        movingObjectToTarget=moving_to_target,
        closeToObject=close_to_object)

    # teleport any hit targets
    rng, target = self._random_target(state.info['rng'])
    target = jp.where(target_hit, target, qp.pos[self.target_idx])
    pos = jp.index_update(qp.pos, self.target_idx, target)
    qp = qp.replace(pos=pos)
    state.info.update(rng=rng)
    return state.replace(qp=qp, obs=obs, reward=reward)

  @property
  def action_size(self) -> int:
    return super().action_size + 3  # 3 extra actions for translating

  def _get_obs(self, qp: brax.QP, info: brax.Info) -> jp.ndarray:
    """Egocentric observation of target, object, and hand."""

    v_inv_rotate = jp.vmap(math.inv_rotate, include=(True, False))

    pos_local = qp.pos - qp.pos[self.palm_idx]
    pos_local = v_inv_rotate(pos_local, qp.rot[self.palm_idx])
    vel_local = v_inv_rotate(qp.vel, qp.rot[self.palm_idx])

    object_local = pos_local[self.object_idx]
    object_local_mag = jp.norm(object_local).reshape(-1)
    object_local_dir = object_local / (1e-6 + object_local_mag)

    # target specific obs
    hand_to_obj = qp.pos[self.object_idx] - qp.pos[self.palm_idx]
    hand_to_obj_mag = jp.norm(hand_to_obj)
    hand_to_obj_dir = hand_to_obj / (1e-6 + hand_to_obj_mag)
    hand_vel = qp.vel[self.hand_idx]
    heading_to_obj = jp.dot(hand_to_obj_dir, hand_vel).reshape(-1)

    target_local = pos_local[self.target_idx]
    target_local_mag = jp.norm(target_local).reshape(-1)
    target_local_dir = target_local / (1e-6 + target_local_mag)

    # object to target
    obj_to_target = qp.pos[self.target_idx] - qp.pos[self.object_idx]
    obj_to_target_mag = jp.norm(obj_to_target).reshape(-1)
    obj_to_target_dir = obj_to_target / (1e-6 + obj_to_target_mag)

    object_vel = qp.vel[self.object_idx]
    obj_heading_to_target = jp.dot(obj_to_target_dir, object_vel).reshape(-1)

    pos_local = pos_local.reshape(-1)
    vel_local = vel_local.reshape(-1)

    contact_mag = jp.sum(jp.square(info.contact.vel), axis=-1)
    contacts = jp.where(contact_mag > 0.00001, 1, 0)

    return jp.concatenate([
        object_local_mag, object_local_dir, target_local_mag, target_local_dir,
        pos_local, vel_local, hand_to_obj, hand_vel, heading_to_obj,
        obj_to_target_mag, obj_to_target_dir, obj_heading_to_target, contacts
    ])

  def _random_target(self, rng: jp.ndarray) -> Tuple[jp.ndarray, jp.ndarray]:
    """Returns new random target locations in a random circle on xz plane."""
    rng, rng1, rng2, rng3 = jp.random_split(rng, 4)
    dist = self.target_radius + self.target_distance * jp.random_uniform(rng1)
    ang = jp.pi * 2. * jp.random_uniform(rng2)
    target_x = dist * jp.cos(ang)
    target_y = dist * jp.sin(ang)
    target_z = self.target_height * jp.random_uniform(rng3)
    target = jp.array([target_x, target_y, target_z]).transpose()
    return rng, target


_SYSTEM_CONFIG = """
  bodies {
    name: "Ground"
    colliders {
      plane {
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
    frozen { all: true }
  }
  bodies {
    name: "Object"
    colliders {
      capsule {
        radius: 1.0
        length: 2.02
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbProximal"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbDistal"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbMiddle"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandPalm"
    colliders {
      capsule {
        radius: 1.5
        length: 3.1
      }
      rotation { x: 120 }
    }
    inertia {
      x: 5000.0350000858306885
      y: 10000.7224998474121094
      z: 10000.7224998474121094
    }
    mass: 1.0
    frozen { all : true }
  }
  bodies {
    name: "HandThumbProximalTwo"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbMiddleTwo"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbDistalTwo"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbProximalThree"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbMiddleThree"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbDistalThree"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbProximalFour"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbMiddleFour"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbDistalFour"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "Target"
    colliders { sphere { radius: 1.1 }}
    frozen { all: true }
  }
  joints {
    name: "HandThumbMiddle_HandThumbDistal"
    angular_damping: 50.0
    parent_offset {
      y: 0.75
    }
    child_offset {
      y: -0.75
    }
    parent: "HandThumbMiddle"
    child: "HandThumbDistal"

    angle_limit { min: -90.0 }
    angle_limit { min: -0.1 max: 0.1 }
    angle_limit { min: -0.1 max: 0.1 }
  }
  joints {
    name: "HandThumbProximal_HandThumbMiddle"
    angular_damping: 50.0
    parent_offset {
      y: 0.75
    }
    child_offset {
      y: -0.75
    }
    parent: "HandThumbProximal"
    child: "HandThumbMiddle"

    angle_limit { min: -90.0 }
    angle_limit { min: -0.1 max: 0.1 }
    angle_limit { min: -0.1 max: 0.1 }
  }
  joints {
    name: "HandPalm_HandThumbProximal"
    angular_damping: 50.0
    parent_offset {
      y: 1.5
      z: -.5
    }
    child_offset {
      y: -0.75
    }
    parent: "HandPalm"
    child: "HandThumbProximal"

    angle_limit { min: -45 max: 45 }
    angle_limit { min: -45 max: 45 }
    angle_limit { min: -0.1 max: 0.1 }
  }
  joints {
    name: "HandThumbMiddleTwo_HandThumbDistalTwo"
    angular_damping: 50.0
    parent_offset {
      y: -0.75
    }
    child_offset {
      y: 0.75
    }
    parent: "HandThumbMiddleTwo"
    child: "HandThumbDistalTwo"

    angle_limit { min: 0 max: 90 }
    angle_limit { min: -0.1 max: 0.1 }
    angle_limit { min: -0.1 max: 0.1 }

  }
  joints {
    name: "HandThumbProximalTwo_HandThumbMiddleTwo"
    angular_damping: 50.0
    parent_offset {
      y: -0.75
    }
    child_offset {
      y: 0.75
    }
    parent: "HandThumbProximalTwo"
    child: "HandThumbMiddleTwo"

    angle_limit { min: 0 max: 90 }
    angle_limit { min: -0.1 max: 0.1 }
    angle_limit { min: -0.1 max: 0.1 }
  }
  joints {
    name: "HandPalm_HandThumbProximalTwo"
    angular_damping: 50.0
    parent_offset {
      y: -1.5
      z: -.5
    }
    child_offset {
      y: 0.75
    }
    parent: "HandPalm"
    child: "HandThumbProximalTwo"

    angle_limit { min: -45 max: 45 }
    angle_limit { min: -45 max: 45 }
    angle_limit { min: -0.1 max: 0.1 }
  }
  joints {
    name: "HandThumbMiddleThree_HandThumbDistalThree"
    angular_damping: 50.0
    parent_offset {
      x: -0.75
    }
    child_offset {
      x: 0.75
    }
    parent: "HandThumbMiddleThree"
    child: "HandThumbDistalThree"

    rotation { z: -90 }
    angle_limit { min: 0 max: 90 }
    angle_limit { min: -0.1 max: 0.1 }
    angle_limit { min: -0.1 max: 0.1 }
  }
  joints {
    name: "HandThumbProximalThree_HandThumbMiddleThree"
    angular_damping: 50.0
    parent_offset {
      x: -0.75
    }
    child_offset {
      x: 0.75
    }
    parent: "HandThumbProximalThree"
    child: "HandThumbMiddleThree"

    rotation { z: -90 }
    angle_limit { min: 0 max: 90 }
    angle_limit { min: -0.1 max: 0.1 }
    angle_limit { min: -0.1 max: 0.1 }
  }
  joints {
    name: "HandPalm_HandThumbProximalThree"
    angular_damping: 50.0
    parent_offset {
      x: -1.5
      z: -.5
    }
    child_offset {
      x: 0.75
    }
    parent: "HandPalm"
    child: "HandThumbProximalThree"

    rotation { z: -90 }
    angle_limit { min: -45 max: 45 }
    angle_limit { min: -45 max: 45 }
    angle_limit { min: -0.1 max: 0.1 }
  }
  joints {
    name: "HandThumbMiddleFour_HandThumbDistalFour"
    angular_damping: 50.0
    parent_offset {
      x: 0.75
    }
    child_offset {
      x: -0.75
    }
    parent: "HandThumbMiddleFour"
    child: "HandThumbDistalFour"

    rotation { z: 90 }
    angle_limit { min: 0 max: 90 }
    angle_limit { min: -0.1 max: 0.1 }
    angle_limit { min: -0.1 max: 0.1 }
  }
  joints {
    name: "HandThumbProximalFour_HandThumbMiddleFour"
    angular_damping: 50.0
    parent_offset {
      x: 0.75
    }
    child_offset {
      x: -0.75
    }
    parent: "HandThumbProximalFour"
    child: "HandThumbMiddleFour"

    rotation { z: 90 }
    angle_limit { min: 0 max: 90 }
    angle_limit { min: -0.1 max: 0.1 }
    angle_limit { min: -0.1 max: 0.1 }
  }
  joints {
    name: "HandPalm_HandThumbProximalFour"
    angular_damping: 50.0
    parent_offset {
      x: 1.5
      z: -.5
    }
    child_offset {
      x: -0.75
    }
    parent: "HandPalm"
    child: "HandThumbProximalFour"

    rotation { x: 180 z: -90 }
    angle_limit { min: -45 max: 45 }
    angle_limit { min: -45 max: 45 }
    angle_limit { min: -0.1 max: 0.1 }
  }
  actuators {
    name: "HandThumbMiddle_HandThumbDistal"
    joint: "HandThumbMiddle_HandThumbDistal"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbProximal_HandThumbMiddle"
    joint: "HandThumbProximal_HandThumbMiddle"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandPalm_HandThumbProximal"
    joint: "HandPalm_HandThumbProximal"
     strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbMiddleTwo_HandThumbDistalTwo"
    joint: "HandThumbMiddleTwo_HandThumbDistalTwo"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbProximalTwo_HandThumbMiddleTwo"
    joint: "HandThumbProximalTwo_HandThumbMiddleTwo"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandPalm_HandThumbProximalTwo"
    joint: "HandPalm_HandThumbProximalTwo"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbMiddleThree_HandThumbDistalThree"
    joint: "HandThumbMiddleThree_HandThumbDistalThree"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbProximalThree_HandThumbMiddleThree"
    joint: "HandThumbProximalThree_HandThumbMiddleThree"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandPalm_HandThumbProximalThree"
    joint: "HandPalm_HandThumbProximalThree"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbMiddleFour_HandThumbDistalFour"
    joint: "HandThumbMiddleFour_HandThumbDistalFour"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbProximalFour_HandThumbMiddleFour"
    joint: "HandThumbProximalFour_HandThumbMiddleFour"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandPalm_HandThumbProximalFour"
    joint: "HandPalm_HandThumbProximalFour"
    strength: 300.0
    angle {}
  }
  friction: 1.0
  gravity {
    z: -9.800000190734863
  }
  angular_damping: -0.05000000074505806
  collide_include {
    first: "Ground"
    second: "Object"
  }
  collide_include {
    first: "HandThumbDistal"
    second: "Object"
  }
  collide_include {
    first: "HandThumbDistalTwo"
    second: "Object"
  }
  collide_include {
    first: "HandThumbDistalThree"
    second: "Object"
  }
  collide_include {
    first: "HandThumbDistalFour"
    second: "Object"
  }
  collide_include {
    first: "HandPalm"
    second: "Object"
  }
  dt: 0.02
  substeps: 4
  dynamics_mode: "pbd"
  """

_SYSTEM_CONFIG_SPRING = """
  bodies {
    name: "Ground"
    colliders {
      plane {
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
    frozen { all: true }
  }
  bodies {
    name: "Object"
    colliders {
      capsule {
        radius: 1.0
        length: 2.02
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbProximal"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbDistal"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbMiddle"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandPalm"
    colliders {
      capsule {
        radius: 1.5
        length: 3.1
      }
      rotation { x: 120 }
    }
    inertia {
      x: 5000.0350000858306885
      y: 10000.7224998474121094
      z: 10000.7224998474121094
    }
    mass: 1.0
    frozen { all : true }
  }
  bodies {
    name: "HandThumbProximalTwo"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbMiddleTwo"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbDistalTwo"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { x: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbProximalThree"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbMiddleThree"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbDistalThree"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbProximalFour"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbMiddleFour"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "HandThumbDistalFour"
    colliders {
      capsule {
        radius: 0.5
        length: 2.0
      }
      rotation { y: 90 }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
  }
  bodies {
    name: "Target"
    colliders { sphere { radius: 1.1 }}
    frozen { all: true }
  }
  joints {
    name: "HandThumbMiddle_HandThumbDistal"
    parent_offset {
      y: 0.75
    }
    child_offset {
      y: -0.75
    }
    parent: "HandThumbMiddle"
    child: "HandThumbDistal"

    stiffness: 5000.0
    angle_limit { min: -90.0 }
    angular_damping: 50.0
  }
  joints {
    name: "HandThumbProximal_HandThumbMiddle"
    parent_offset {
      y: 0.75
    }
    child_offset {
      y: -0.75
    }
    parent: "HandThumbProximal"
    child: "HandThumbMiddle"

    stiffness: 5000.0
    angle_limit { min: -90.0 }
    angular_damping: 50.0
  }
  joints {
    name: "HandPalm_HandThumbProximal"
    parent_offset {
      y: 1.5
      z: -.5
    }
    child_offset {
      y: -0.75
    }
    parent: "HandPalm"
    child: "HandThumbProximal"

    stiffness: 5000.0
    angle_limit { min: -45 max: 45 }
    angle_limit { min: -45 max: 45 }
    angular_damping: 50.0
  }
  joints {
    name: "HandThumbMiddleTwo_HandThumbDistalTwo"
    parent_offset {
      y: -0.75
    }
    child_offset {
      y: 0.75
    }
    parent: "HandThumbMiddleTwo"
    child: "HandThumbDistalTwo"

    stiffness: 5000.0
    angle_limit { min: 0 max: 90 }
    angular_damping: 50.0

  }
  joints {
    name: "HandThumbProximalTwo_HandThumbMiddleTwo"
    parent_offset {
      y: -0.75
    }
    child_offset {
      y: 0.75
    }
    parent: "HandThumbProximalTwo"
    child: "HandThumbMiddleTwo"

    stiffness: 5000.0
    angle_limit { min: 0 max: 90 }
    angular_damping: 50.0
  }
  joints {
    name: "HandPalm_HandThumbProximalTwo"
    parent_offset {
      y: -1.5
      z: -.5
    }
    child_offset {
      y: 0.75
    }
    parent: "HandPalm"
    child: "HandThumbProximalTwo"

    stiffness: 5000.0
    angle_limit { min: -45 max: 45 }
    angle_limit { min: -45 max: 45 }
    angular_damping: 50.0
  }
  joints {
    name: "HandThumbMiddleThree_HandThumbDistalThree"
    parent_offset {
      x: -0.75
    }
    child_offset {
      x: 0.75
    }
    parent: "HandThumbMiddleThree"
    child: "HandThumbDistalThree"

    stiffness: 5000.0
    rotation { z: -90 }
    angle_limit { min: 0 max: 90 }
    angular_damping: 50.0
  }
  joints {
    name: "HandThumbProximalThree_HandThumbMiddleThree"
    parent_offset {
      x: -0.75
    }
    child_offset {
      x: 0.75
    }
    parent: "HandThumbProximalThree"
    child: "HandThumbMiddleThree"

    stiffness: 5000.0
    rotation { z: -90 }
    angle_limit { min: 0 max: 90 }
    angular_damping: 50.0
  }
  joints {
    name: "HandPalm_HandThumbProximalThree"
    parent_offset {
      x: -1.5
      z: -.5
    }
    child_offset {
      x: 0.75
    }
    parent: "HandPalm"
    child: "HandThumbProximalThree"

    stiffness: 5000.0
    rotation { z: -90 }
    angle_limit { min: -45 max: 45 }
    angle_limit { min: -45 max: 45 }
    angular_damping: 50.0
  }
  joints {
    name: "HandThumbMiddleFour_HandThumbDistalFour"
    parent_offset {
      x: 0.75
    }
    child_offset {
      x: -0.75
    }
    parent: "HandThumbMiddleFour"
    child: "HandThumbDistalFour"

    stiffness: 5000.0
    rotation { z: 90 }
    angle_limit { min: 0 max: 90 }
    angular_damping: 50.0
  }
  joints {
    name: "HandThumbProximalFour_HandThumbMiddleFour"
    parent_offset {
      x: 0.75
    }
    child_offset {
      x: -0.75
    }
    parent: "HandThumbProximalFour"
    child: "HandThumbMiddleFour"

    stiffness: 5000.0
    rotation { z: 90 }
    angle_limit { min: 0 max: 90 }
    angular_damping: 50.0
  }
  joints {
    name: "HandPalm_HandThumbProximalFour"
    parent_offset {
      x: 1.5
      z: -.5
    }
    child_offset {
      x: -0.75
    }
    parent: "HandPalm"
    child: "HandThumbProximalFour"

    stiffness: 5000.0
    rotation { x: 180 z: -90 }
    angle_limit { min: -45 max: 45 }
    angle_limit { min: -45 max: 45 }
    angular_damping: 50.0
  }
  actuators {
    name: "HandThumbMiddle_HandThumbDistal"
    joint: "HandThumbMiddle_HandThumbDistal"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbProximal_HandThumbMiddle"
    joint: "HandThumbProximal_HandThumbMiddle"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandPalm_HandThumbProximal"
    joint: "HandPalm_HandThumbProximal"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbMiddleTwo_HandThumbDistalTwo"
    joint: "HandThumbMiddleTwo_HandThumbDistalTwo"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbProximalTwo_HandThumbMiddleTwo"
    joint: "HandThumbProximalTwo_HandThumbMiddleTwo"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandPalm_HandThumbProximalTwo"
    joint: "HandPalm_HandThumbProximalTwo"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbMiddleThree_HandThumbDistalThree"
    joint: "HandThumbMiddleThree_HandThumbDistalThree"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbProximalThree_HandThumbMiddleThree"
    joint: "HandThumbProximalThree_HandThumbMiddleThree"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandPalm_HandThumbProximalThree"
    joint: "HandPalm_HandThumbProximalThree"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbMiddleFour_HandThumbDistalFour"
    joint: "HandThumbMiddleFour_HandThumbDistalFour"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandThumbProximalFour_HandThumbMiddleFour"
    joint: "HandThumbProximalFour_HandThumbMiddleFour"
    strength: 300.0
    angle {}
  }
  actuators {
    name: "HandPalm_HandThumbProximalFour"
    joint: "HandPalm_HandThumbProximalFour"
    strength: 300.0
    angle {}
  }
  friction: 0.77459666924
  gravity {
    z: -9.800000190734863
  }
  angular_damping: -0.05000000074505806
  baumgarte_erp: 0.15000000149011612
  collide_include {
    first: "Ground"
    second: "Object"
  }
  collide_include {
    first: "HandThumbDistal"
    second: "Object"
  }
  collide_include {
    first: "HandThumbDistalTwo"
    second: "Object"
  }
  collide_include {
    first: "HandThumbDistalThree"
    second: "Object"
  }
  collide_include {
    first: "HandThumbDistalFour"
    second: "Object"
  }
  collide_include {
    first: "HandPalm"
    second: "Object"
  }
  dt: 0.02
  substeps: 4
  dynamics_mode: "legacy_spring"
  """




# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Procedural ant."""
from brax.experimental.composer.components.ant import DEFAULT_OBSERVERS
from brax.experimental.composer.components.ant import ROOT
from brax.experimental.composer.components.ant import term_fn
import numpy as np


def generate_ant_config_with_n_legs(n):
  """Generate info for n-legged ant."""

  def template_leg(theta, ind):
    tmp = f"""
      bodies {{
      name: "Aux 1_{str(ind)}"
      colliders {{
        rotation {{ x: 90 y: -90 }}
        capsule {{
          radius: 0.08
          length: 0.4428427219390869
        }}
      }}
      inertia {{ x: 1.0 y: 1.0 z: 1.0 }}
      mass: 1
    }}
    bodies {{
      name: "$ Body 4_{str(ind)}"
      colliders {{
        rotation {{ x: 90 y: -90 }}
        capsule {{
          radius: 0.08
          length: 0.7256854176521301
          end: -1
        }}
      }}
      inertia {{ x: 1.0 y: 1.0 z: 1.0 }}
      mass: 1
    }}
      joints {{
        name: "{ROOT}_Aux 1_{str(ind)}"
        parent_offset {{ x: {((0.4428427219390869/2.)+.08)*np.cos(theta)} y: {((0.4428427219390869/2.)+.08)*np.sin(theta)} }}
        child_offset {{ }}
        parent: "{ROOT}"
        child: "Aux 1_{str(ind)}"
        stiffness: 5000.0
        angular_damping: 35
        angle_limit {{ min: -30.0 max: 30.0 }}
        rotation {{ y: -90 }}
        reference_rotation {{ z: {theta*180/np.pi} }}
      }}
      joints {{
      name: "Aux 1_$ Body 4_{str(ind)}"
      parent_offset {{ x: {0.4428427219390869/2. - .08}  }}
      child_offset {{ x:{-0.7256854176521301/2. + .08}  }}
      parent: "Aux 1_{str(ind)}"
      child: "$ Body 4_{str(ind)}"
      stiffness: 5000.0
      angular_damping: 35
      rotation: {{ z: 90 }}
      angle_limit {{
        min: 30.0
        max: 70.0
      }}
    }}
    actuators {{
      name: "{ROOT}_Aux 1_{str(ind)}"
      joint: "{ROOT}_Aux 1_{str(ind)}"
      strength: 350.0
      torque {{}}
    }}
    actuators {{
      name: "Aux 1_$ Body 4_{str(ind)}"
      joint: "Aux 1_$ Body 4_{str(ind)}"
      strength: 350.0
      torque {{}}
    }}
    """
    collides = (f'Aux 1_{str(ind)}', f'$ Body 4_{str(ind)}')
    return tmp, collides

  base_config = f"""
    bodies {{
      name: "{ROOT}"
      colliders {{
        capsule {{
          radius: 0.25
          length: 0.5
          end: 1
        }}
      }}
      inertia {{ x: 1.0 y: 1.0 z: 1.0 }}
      mass: 10
    }}
    """
  collides = (ROOT,)
  for i in range(n):
    config_i, collides_i = template_leg((1. * i / n) * 2 * np.pi, i)
    base_config += config_i
    collides += collides_i

  return base_config, collides


def get_specs(num_legs: int = 10):
  message_str, collides = generate_ant_config_with_n_legs(num_legs)
  return dict(
      message_str=message_str,
      collides=collides,
      root=ROOT,
      term_fn=term_fn,
      observers=DEFAULT_OBSERVERS)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains a reacher to reach a target.

Based on the OpenAI Gym MuJoCo Reacher environment.
"""

from typing import Tuple

import brax
from brax import jumpy as jp
from brax.envs import env


class Reacher(env.Env):



  """
  ### Description

  "Reacher" is a two-jointed robot arm. The goal is to move the robot's end
  effector (called *fingertip*) close to a target that is spawned at a random
  position.

  ### Action Space

  The action space is a `Box(-1, 1, (2,), float32)`. An action `(a, b)`
  represents the torques applied at the hinge joints.

  | Num | Action                                                                          | Control Min | Control Max | Name (in corresponding config) | Joint | Unit         |
  |-----|---------------------------------------------------------------------------------|-------------|-------------|--------------------------------|-------|--------------|
  | 0   | Torque applied at the first hinge (connecting the link to the point of fixture) | -1          | 1           | joint0                         | hinge | torque (N m) |
  | 1   | Torque applied at the second hinge (connecting the two links)                   | -1          | 1           | joint1                         | hinge | torque (N m) |

  ### Observation Space

  Observations consist of:

  - The cosine of the angles of the two arms
  - The sine of the angles of the two arms
  - The coordinates of the target
  - The angular velocities of the arms
  - The vector between the target and the reacher's fingertip (3 dimensional
    with the last element being 0)

  The observation is a `ndarray` with shape `(11,)` where the elements
  correspond to the following:

  | Num | Observation                                                                                    | Min  | Max | Name (in corresponding config) | Joint | Unit                     |
  |-----|------------------------------------------------------------------------------------------------|------|-----|--------------------------------|-------|--------------------------|
  | 0   | cosine of the angle of the first arm                                                           | -Inf | Inf | cos(joint0)                    | hinge | unitless                 |
  | 1   | cosine of the angle of the second arm                                                          | -Inf | Inf | cos(joint1)                    | hinge | unitless                 |
  | 2   | sine of the angle of the first arm                                                             | -Inf | Inf | cos(joint0)                    | hinge | unitless                 |
  | 3   | sine of the angle of the second arm                                                            | -Inf | Inf | cos(joint1)                    | hinge | unitless                 |
  | 4   | x-coordinate of the target                                                                     | -Inf | Inf | target_x                       | slide | position (m)             |
  | 5   | y-coordinate of the target                                                                     | -Inf | Inf | target_y                       | slide | position (m)             |
  | 6   | angular velocity of the first arm                                                              | -Inf | Inf | joint0                         | hinge | angular velocity (rad/s) |
  | 7   | angular velocity of the second arm                                                             | -Inf | Inf | joint1                         | hinge | angular velocity (rad/s) |
  | 8   | x-value of position_fingertip - position_target                                                | -Inf | Inf | NA                             | slide | position (m)             |
  | 9   | y-value of position_fingertip - position_target                                                | -Inf | Inf | NA                             | slide | position (m)             |
  | 10  | z-value of position_fingertip - position_target (0 since reacher is 2d and z is same for both) | -Inf | Inf | NA                             | slide | position (m)             |

  ### Rewards

  The reward consists of two parts:

  - *reward_dist*: This reward is a measure of how far the *fingertip*
    of the reacher (the unattached end) is from the target, with a more negative
    value assigned for when the reacher's *fingertip* is further away from the
    target. It is calculated as the negative vector norm of (position of
    the fingertip - position of target), or *-norm("fingertip" - "target")*.
  - *reward_ctrl*: A negative reward for penalising the walker if it takes
    actions that are too large. It is measured as the negative squared
    Euclidean norm of the action, i.e. as *- sum(action<sup>2</sup>)*.

  Unlike other environments, Reacher does not allow you to specify weights for
  the individual reward terms. However, `info` does contain the keys
  *reward_dist* and *reward_ctrl*. Thus, if you'd like to weight the terms, you
  should create a wrapper that computes the weighted reward from `info`.

  ### Starting State

  All observations start in state (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
  0.0, 0.0) with a noise added for stochasticity. A uniform noise in the range
  [-0.1, 0.1] is added to the positional attributes, while the target position
  is selected uniformly at random in a disk of radius 0.2 around the origin.

  Independent, uniform noise in the range of [-0.005, 0.005] is added to the
  velocities, and the last element ("fingertip" - "target") is calculated at the
  end once everything is set.

  The default setting has a *dt = 0.02*

  ### Episode Termination

  The episode terminates when any of the following happens:

  1. The episode duration reaches a 1000 timesteps

  ### Arguments

  No additional arguments are currently supported (in v2 and lower), but
  modifications can be made to the XML file in the assets folder (or by changing
  the path to a modified XML file in another folder)..

  ```
  env = gym.make('Reacher-v2')
  ```

  There is no v3 for Reacher, unlike the robot environments where a v3 and
  beyond take gym.make kwargs such as ctrl_cost_weight, reset_noise_scale etc.

  There is a v4 version that uses the mujoco-bindings

  ```
  env = gym.make('Reacher-v4')
  ```

  And a v5 version that uses Brax:

  ```
  env = gym.make('Reacher-v5')
  ```

  ### Version History

  * v5: ported to Brax.
  * v4: all mujoco environments now use the mujoco bindings in mujoco>=2.1.3
  * v2: All continuous control environments now use mujoco_py >= 1.50
  * v1: max_time_steps raised to 1000 for robot based tasks (not including
    reacher, which has a max_time_steps of 50). Added reward_threshold to
    environments.
  * v0: Initial versions release (1.0.0)
  """


  def __init__(self, legacy_spring=False, **kwargs):
    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
    super().__init__(config=config, **kwargs)
    self._target_idx = self.sys.body.index['target']
    self._arm_idx = self.sys.body.index['body1']

  def reset(self, rng: jp.ndarray) -> env.State:
    rng, rng1, rng2 = jp.random_split(rng, 3)

    qpos = self.sys.default_angle() + jp.random_uniform(
        rng1, (self.sys.num_joint_dof,), -.1, .1)
    qvel = jp.random_uniform(rng2, (self.sys.num_joint_dof,), -.005, .005)

    qp = self.sys.default_qp(joint_angle=qpos, joint_velocity=qvel)
    _, target = self._random_target(rng)
    pos = jp.index_update(qp.pos, self._target_idx, target)

    qp = qp.replace(pos=pos)
    obs = self._get_obs(qp, self.sys.info(qp))
    reward, done, zero = jp.zeros(3)
    metrics = {
        'reward_dist': zero,
        'reward_ctrl': zero,
    }
    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    qp, info = self.sys.step(state.qp, action)
    obs = self._get_obs(qp, info)

    # vector from tip to target is last 3 entries of obs vector
    reward_dist = -jp.norm(obs[-3:])
    reward_ctrl = -jp.square(action).sum()
    reward = reward_dist + reward_ctrl

    state.metrics.update(
        reward_dist=reward_dist,
        reward_ctrl=reward_ctrl,
    )

    return state.replace(qp=qp, obs=obs, reward=reward)

  def _get_obs(self, qp: brax.QP, info: brax.Info) -> jp.ndarray:
    """Egocentric observation of target and arm body."""
    joint_angle, _ = self.sys.joints[0].angle_vel(qp)

    # qpos:
    # x,y coord of target
    qpos = [qp.pos[self._target_idx, :2]]

    # dist to target and speed of tip
    arm_qps = jp.take(qp, jp.array(self._arm_idx))
    tip_pos, tip_vel = arm_qps.to_world(jp.array([0.11, 0., 0.]))
    tip_to_target = [tip_pos - qp.pos[self._target_idx]]
    cos_sin_angle = [jp.cos(joint_angle), jp.sin(joint_angle)]

    # qvel:
    # velocity of tip
    qvel = [tip_vel[:2]]

    return jp.concatenate(cos_sin_angle + qpos + qvel + tip_to_target)

  def _random_target(self, rng: jp.ndarray) -> Tuple[jp.ndarray, jp.ndarray]:
    """Returns a target location in a random circle slightly above xy plane."""
    rng, rng1, rng2 = jp.random_split(rng, 3)
    dist = .2 * jp.random_uniform(rng1)
    ang = jp.pi * 2. * jp.random_uniform(rng2)
    target_x = dist * jp.cos(ang)
    target_y = dist * jp.sin(ang)
    target_z = .01
    target = jp.array([target_x, target_y, target_z]).transpose()
    return rng, target


_SYSTEM_CONFIG = """
  bodies {
    name: "ground"
    colliders {
      plane {
      }
    }
    mass: 1.0
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    frozen {
      all: true
    }
  }
  bodies {
    name: "body0"
    colliders {
      position {
        x: 0.05
      }
      rotation {
        y: 90.0
      }
      capsule {
        radius: 0.01
        length: 0.12
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.035604715
  }
  bodies {
    name: "body1"
    colliders {
      position {
        x: 0.05
      }
      rotation {
        y: 90.0
      }
      capsule {
        radius: 0.01
        length: 0.12
      }
    }
    colliders {
      position { x: .11 }
      sphere {
        radius: 0.01
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.035604715
  }
  bodies {
    name: "target"
    colliders {
      position {
      }
      sphere {
        radius: 0.009
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
    frozen { all: true }
  }
  joints {
    name: "joint0"
    parent: "ground"
    child: "body0"
    parent_offset {
      z: 0.01
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angle_limit {
        min: -360
        max: 360
      }
  }
  joints {
    name: "joint1"
    parent: "body0"
    child: "body1"
    parent_offset {
      x: 0.1
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angle_limit {
      min: -360
      max: 360
    }
  }
  actuators {
    name: "joint0"
    joint: "joint0"
    strength: 25.0
    torque {
    }
  }
  actuators {
    name: "joint1"
    joint: "joint1"
    strength: 25.0
    torque {
    }
  }
  collide_include {
  }
  gravity {
    z: -9.81
  }
  dt: 0.02
  substeps: 4
  frozen {
    position {
      z: 1.0
    }
    rotation {
      x: 1.0
      y: 1.0
    }
  }
  dynamics_mode: "pbd"
  """


_SYSTEM_CONFIG_SPRING = """
  bodies {
    name: "ground"
    colliders {
      plane {
      }
    }
    mass: 1.0
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    frozen {
      all: true
    }
  }
  bodies {
    name: "body0"
    colliders {
      position {
        x: 0.05
      }
      rotation {
        y: 90.0
      }
      capsule {
        radius: 0.01
        length: 0.12
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.035604715
  }
  bodies {
    name: "body1"
    colliders {
      position {
        x: 0.05
      }
      rotation {
        y: 90.0
      }
      capsule {
        radius: 0.01
        length: 0.12
      }
    }
    colliders {
      position { x: .11 }
      sphere {
        radius: 0.01
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 0.035604715
  }
  bodies {
    name: "target"
    colliders {
      position {
      }
      sphere {
        radius: 0.009
      }
    }
    inertia {
      x: 1.0
      y: 1.0
      z: 1.0
    }
    mass: 1.0
    frozen { all: true }
  }
  joints {
    name: "joint0"
    stiffness: 100.0
    parent: "ground"
    child: "body0"
    parent_offset {
      z: 0.01
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angle_limit {
        min: -360
        max: 360
      }
    limit_strength: 0.0
    spring_damping: 3.0
  }
  joints {
    name: "joint1"
    stiffness: 100.0
    parent: "body0"
    child: "body1"
    parent_offset {
      x: 0.1
    }
    child_offset {
    }
    rotation {
      y: -90.0
    }
    angle_limit {
      min: -360
      max: 360
    }
    limit_strength: 0.0
    spring_damping: 3.0
  }
  actuators {
    name: "joint0"
    joint: "joint0"
    strength: 25.0
    torque {
    }
  }
  actuators {
    name: "joint1"
    joint: "joint1"
    strength: 25.0
    torque {
    }
  }
  collide_include {
  }
  gravity {
    z: -9.81
  }
  baumgarte_erp: 0.1
  dt: 0.02
  substeps: 4
  frozen {
    position {
      z: 1.0
    }
    rotation {
      x: 1.0
      y: 1.0
    }
  }
  dynamics_mode: "legacy_spring"
  """
